import copy

import numpy as np
from aiondao import config
from aiondao.agents.commons import TokenBatch
from aiondao.agents.members import Participant, Proposal, ProposalStatus
from aiondao.libs.conviction import trigger_threshold
from aiondao.utils.networking import (
    add_participant,
    add_proposal,
    calc_median_affinity,
    calc_total_conviction,
    calc_total_funds_requested,
    find_in_edges_of_type_for_proposal,
    get_edges_by_type,
    get_participants,
    get_proposals,
    get_proposals_by_participant_and_status,
)
from aiondao.utils import _utilz
from auto_all import end_all, start_all
from loguru import logger

start_all(globals())


class GenerateNewParticipant:
    @staticmethod
    def p_randomly(params, step, sL, s, **kwargs):
        commons = s["commons"]
        sentiment = s["sentiment"]
        timestep = s["timestep"]
        probability_func = params["probability_func"]
        exponential_func = params["exponential_func"]
        speculation_days = params["speculation_days"]
        multiplier_new_participants = params["multiplier_new_participants"]

        ans = {"new_participant_investment": None, "new_participant_tokens": None}
        dict_ans = {}
        if timestep < speculation_days:
            # If in speculation period, the arrival rate is higher
            arrival_rate = 0.5 + 0.5 * sentiment
            multiplier = multiplier_new_participants
            max_new_participants = config.max_new_participants * multiplier
        else:
            arrival_rate = (1 + sentiment) / config.arrival_rate_denominator
            max_new_participants = config.max_new_participants

        for i in range(max_new_participants):
            if probability_func(arrival_rate):
                # Here we randomly generate each participant's post-Hatch
                # investment, in DAI/USD.
                #
                # expon.rvs() arguments:
                #
                # loc is the minimum number, so if loc=100, there will be no
                # investments < 100
                #
                # scale is the standard deviation, so if scale=2, investments will
                # be around 0-12 DAI or even 15, if scale=100, the investments will be
                # around 0-600 DAI.
                ans["new_participant_investment"] = exponential_func(
                    loc=config.investment_new_participant_min,
                    scale=config.investment_new_participant_stdev,
                )
                ans["new_participant_tokens"] = commons.dai_to_tokens(
                    ans["new_participant_investment"]
                )
                dict_ans[i] = ans
        return dict_ans

    @staticmethod
    def su_add_to_network(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        probability_func = params["probability_func"]
        exponential_func = params["exponential_func"]
        random_number_func = params["random_number_func"]

        for i in _input:
            ans = _input[i]
            if ans != 0:
                network, i = add_participant(
                    network,
                    Participant(
                        holdings=TokenBatch(0, ans["new_participant_tokens"]),
                        prob_fn=probability_func,
                        random_number_func=random_number_func,
                    ),
                    exponential_func,
                    random_number_func,
                )

                if params.get("debug"):
                    print(
                        "GenerateNewParticipant: A new Participant {} invested {}DAI for {} tokens".format(
                            i,
                            ans["new_participant_investment"],
                            ans["new_participant_tokens"],
                        )
                    )
        return "network", network

    @staticmethod
    def su_add_investment_to_commons(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]
        for i in _input:
            ans = _input[i]
            if ans != 0:
                tokens, realized_price = commons.deposit(
                    ans["new_participant_investment"]
                )
        return "commons", commons

    @staticmethod
    def su_update_participants_token_batch_age(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        participants = get_participants(network)
        for i, participant in participants:
            participant.update_token_batch_age()

        return "network", network


class GenerateNewProposal:
    @staticmethod
    def p_randomly(params, step, sL, s, **kwargs):
        """
        Randomly picks a Participant from the network and asks him if he wants
        to create a Proposal.
        """
        funding_pool = s["funding_pool"]
        network = s["network"]
        choice_func = params["choice_func"]

        participants = get_participants(network)
        participants_dict = dict(participants)
        idx = choice_func(list(participants_dict))
        participant = participants_dict[idx]

        wants_to_create_proposal = participant.create_proposal(
            calc_total_funds_requested(network),
            calc_median_affinity(network),
            funding_pool,
        )

        return {
            "new_proposal": wants_to_create_proposal,
            "proposed_by_participant": idx,
        }

    @staticmethod
    def su_add_to_network(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        funding_pool = s["funding_pool"]
        token_supply = s["token_supply"]
        gamma_func = params["gamma_func"]
        random_number_func = params["random_number_func"]
        if _input["new_proposal"]:
            proposed_by = _input["proposed_by_participant"]
            # Create the Proposal and add it to the network.
            rescale = funding_pool * config.scale_factor
            r_rv = gamma_func(
                config.funds_requested_alpha,
                loc=config.funds_requested_min,
                scale=rescale,
            )
            proposal = Proposal(
                funds_requested=r_rv,
                trigger=trigger_threshold(
                    r_rv, funding_pool, token_supply, params["max_proposal_request"]
                ),
            )
            network, proposal_idx = add_proposal(
                network, proposal, proposed_by, random_number_func
            )

            if params.get("debug"):
                print(
                    "GenerateNewProposal: Participant {} created Proposal {}".format(
                        _input["proposed_by_participant"], proposal_idx
                    )
                )
        return "network", network


class GenerateNewFunding:
    @staticmethod
    def p_exit_tribute_of_average_speculator_position_size(
        params, step, sL, s, **kwargs
    ):
        """
        This policy needs Commons.exit_tribute to NOT be 0!

        TODO: buy tokens and sell them immediately within the same simulation
        step, assuming a certain position size.
        """
        exponential_func = params["exponential_func"]
        exits = [
            exponential_func(
                loc=config.speculator_position_size_min,
                scale=config.speculator_position_size_stdev,
            )
            for i in range(config.speculators)
        ]
        commons = s["commons"]
        funding = sum(exits) * commons.exit_tribute
        return {"funding": funding}

    @staticmethod
    def su_add_funding(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]
        if _input["funding"]:
            commons._funding_pool += _input["funding"]
        return "commons", commons


class ActiveProposals:
    @staticmethod
    def p_influenced_by_grant_size(params, step, sL, s, **kwargs):
        network = s["network"]
        probability_func = params["probability_func"]

        active_proposals = get_proposals(network, status=ProposalStatus.ACTIVE)
        proposals_that_will_fail = []
        proposals_that_will_succeed = []

        for idx, proposal in active_proposals:
            r_failure = 1 / (
                config.base_failure_rate + np.log(proposal.funds_requested)
            )
            r_success = 1 / (
                config.base_success_rate + np.log(proposal.funds_requested)
            )
            if probability_func(r_failure):
                proposals_that_will_fail.append(idx)
            elif probability_func(r_success):
                proposals_that_will_succeed.append(idx)
        return {
            "failed": proposals_that_will_fail,
            "succeeded": proposals_that_will_succeed,
        }

    @staticmethod
    def su_set_proposal_status(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        for idx in _input["failed"]:
            network.nodes[idx]["item"].status = ProposalStatus.FAILED

        for idx in _input["succeeded"]:
            network.nodes[idx]["item"].status = ProposalStatus.COMPLETED

        return "network", network


class ProposalFunding:
    @staticmethod
    def p_compare_conviction_and_threshold(params, step, sL, s, **kwargs):
        """
        This policy simply goes through the Proposals to see if their thresholds
        are smaller than their gathered conviction
        """
        network = s["network"]
        funding_pool = s["funding_pool"]
        token_supply = s["token_supply"]

        proposals_w_enough_conviction = []
        proposals = get_proposals(network, status=ProposalStatus.CANDIDATE)
        for idx, proposal in proposals:
            total_conviction = calc_total_conviction(network, idx)
            proposal.conviction = total_conviction
            res = proposal.has_enough_conviction(
                funding_pool, token_supply, params["max_proposal_request"]
            )

            if params.get("debug"):
                print(
                    "ProposalFunding: Proposal {} has {} conviction, and needs {} to pass".format(
                        idx, proposal.conviction, proposal.trigger
                    )
                )
            if res:
                proposals_w_enough_conviction.append(idx)

        return {"proposal_idxs_with_enough_conviction": proposals_w_enough_conviction}

    @staticmethod
    def su_make_proposal_active(params, step, sL, s, _input, **kwargs):
        network = s["network"]

        for idx in _input["proposal_idxs_with_enough_conviction"]:
            network.nodes[idx]["item"].status = ProposalStatus.ACTIVE

        return "network", network

    @staticmethod
    def su_deduct_funds_from_funding_pool(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]
        network = s["network"]
        for idx in _input["proposal_idxs_with_enough_conviction"]:
            funds = network.nodes[idx]["item"].funds_requested
            if params.get("debug"):
                print(
                    "ProposalFunding: Proposal {} passed! deducting {} from Commons funding pool".format(
                        idx, funds
                    )
                )
            commons.spend(funds)
        return "commons", commons

    @staticmethod
    def su_update_age_and_conviction_thresholds(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        funding_pool = s["funding_pool"]
        token_supply = s["token_supply"]

        proposals = get_proposals(s["network"], status=ProposalStatus.CANDIDATE)
        for _, proposal in proposals:
            proposal.update_age()
            proposal.update_threshold(
                funding_pool,
                token_supply,
                max_proposal_request=params["max_proposal_request"],
            )

        return "network", network

    @staticmethod
    def su_calculate_conviction(params, step, sL, s, _input, **kwargs):
        """
        Actually calculates the conviction. This function should only run ONCE
        per timestep/iteration!
        """
        network = s["network"]
        alpha = params["alpha_days_to_80p_of_max_voting_weight"]

        support_edges = get_edges_by_type(network, "support")
        for i, j in support_edges:
            proposal_status = network.nodes[j]["item"].status
            if proposal_status == ProposalStatus.CANDIDATE:
                edge = network.edges[i, j]
                prior_conviction = edge["support"].conviction
                current_tokens = edge["support"].tokens
                current_conviction = current_tokens + alpha * prior_conviction
                support_edge = edge["support"]
                support_edge.conviction = current_conviction
                edge["support"] = support_edge

                if params.get("debug") and s["timestep"] == 1:
                    print(
                        "ProposalFunding: Participant {} initially has staked {} tokens on Proposal {}, which will result in {} conviction in the next timestep".format(
                            i, current_tokens, j, edge["support"].conviction
                        )
                    )

        return "network", network


class ParticipantVoting:
    @staticmethod
    def p_participant_votes_on_proposal_according_to_affinity(
        params, step, sL, s, **kwargs
    ):
        """
        This policy collects data from the DiGraph to tell the Participant class
        which candidate proposals it supports, and
        Participant.vote_on_candidate_proposals() will decide which proposals it
        will take action on.

        Then, Participant.stake_across_all_supported_proposals() will tell us
        how much it will stake on each of them.
        """
        network = s["network"]
        participants = get_participants(network)

        participants_stakes = {}
        for participant_idx, participant in participants:
            proposal_idx_affinity = {}  # {4: 0.9, 5: 0.9}
            candidate_proposals = get_proposals_by_participant_and_status(
                network,
                participant_idx=participant_idx,
                status_filter=[ProposalStatus.CANDIDATE],
            )
            for proposal_idx, proposal in candidate_proposals.items():
                proposal_idx_affinity[proposal_idx] = proposal["support"].affinity
            proposals_that_participant_cares_enough_to_vote_on = (
                participant.vote_on_candidate_proposals(proposal_idx_affinity)
            )

            stake_across_all_supported_proposals_input = []
            for (
                proposal_idx,
                affinity,
            ) in proposals_that_participant_cares_enough_to_vote_on.items():
                stake_across_all_supported_proposals_input.append(
                    (affinity, proposal_idx)
                )
            stakes = participant.stake_across_all_supported_proposals(
                stake_across_all_supported_proposals_input
            )

            participants_stakes[participant_idx] = stakes

            if params.get("debug"):
                if proposals_that_participant_cares_enough_to_vote_on:
                    print(
                        "ParticipantVoting: Participant {} was given Proposals with corresponding affinities {} and he decided to vote on {}, distributing his tokens thusly {}".format(
                            participant_idx,
                            proposal_idx_affinity,
                            proposals_that_participant_cares_enough_to_vote_on,
                            stakes,
                        )
                    )

        return {"participants_stake_on_proposals": participants_stakes}

    @staticmethod
    def su_update_participants_votes(params, step, sL, s, _input, **kwargs):
        """
        Simply update the support edges with the new amount of tokens the
        Participant has staked on the Proposal. Leave the conviction calculation
        to another state update function.
        """
        network = s["network"]
        _input = _input["participants_stake_on_proposals"]

        for participant_idx, v in _input.items():
            for proposal_idx, tokens_staked in v.items():
                # I assume there is no longer a need to recalculate a la
                # https://github.com/randomshinichi/conviction/blob/9d1bc9513475dc30d33e3232b385234d0295d361/conviction_system_logic3.py#L510
                # because the tokens were already distributed proportionally to
                # the affinities in
                # p_participant_votes_on_proposal_according_to_affinity()
                # Also, do not recalculate conviction here. Leave that to ProposalFunding.su_calculate_conviction()
                start_support = network[participant_idx][proposal_idx]["support"]
                network[participant_idx][proposal_idx][
                    "support"
                ] = _utilz.base_model_copy(start_support, dict(tokens=tokens_staked))

        return "network", network


class ParticipantBuysTokens:
    """
    When implementing buying tokens, Participants may either buy in bulk with no
    slippage, or endure slippage while buying. Neither is realistic but I'd say
    no slippage is closer to what Participants will experience in real life.

    When implementing selling tokens, slippage is more likely to change the
    decision. In any case, let us implement no slippage first since it is easier
    to reason with.

    Also, buying in bulk is easier to implement in cadCAD because if the state
    update function changes the Commons object each time a Participant buys
    in/sells out, then it would have to update the network and commons object in
    the same function, which is not allowed in cadCAD.
    """

    @staticmethod
    def p_decide_to_buy_tokens_bulk(params, step, sL, s, **kwargs):
        network = s["network"]
        commons = s["commons"]
        participants = get_participants(network)
        ans = {}
        total_dai = 0
        for i, participant in participants:
            # If a participant decides to buy, it will be specified in units of DAI.
            # If a participant decides to sell, it will be specified in units of tokens.
            x = participant.buy()
            if x > 0:
                total_dai += x
                ans[i] = x

        # Now that we have the sum of DAI, ask the Commons object how many
        # tokens this would be minted as a result. This will be inaccurate due
        # to slippage, and we need the result of this policy to be final to
        # avoid chaining 2 state update functions, so we run the deposit on a
        # throwaway copy of Commons
        if total_dai == 0:
            if params.get("debug"):
                print(
                    "ParticipantBuysTokens: No Participants bought tokens in timestep {}".format(
                        step
                    )
                )
            return {
                "participant_decisions": ans,
                "total_dai": 0,
                "tokens": 0,
                "token_price": 0,
                "final_token_distribution": {},
            }

        else:
            commons2 = copy.copy(commons)
            tokens, token_price = commons2.deposit(total_dai)

            final_token_distribution = {}
            for i in ans:
                final_token_distribution[i] = ans[i] / total_dai

            if params.get("debug"):
                print(
                    "ParticipantBuysTokens: These Participants have decided to buy tokens with this amount of DAI: {}".format(
                        ans
                    )
                )
                print(
                    "ParticipantBuysTokens: A total of {} DAI will be deposited. {} tokens should be minted as a result at a price of {} DAI/token".format(
                        total_dai, tokens, token_price
                    )
                )

            return {
                "participant_decisions": ans,
                "total_dai": total_dai,
                "tokens": tokens,
                "token_price": token_price,
                "final_token_distribution": final_token_distribution,
            }

    @staticmethod
    def su_buy_participants_tokens(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]

        if _input["total_dai"] > 0:
            tokens, realized_price = commons.deposit(_input["total_dai"])
            if _input["tokens"] != tokens or _input["token_price"] != realized_price:
                raise Exception(
                    "ParticipantBuysTokens: {} tokens were minted at a price of {} (expected: {} with price {})".format(
                        tokens, realized_price, _input["tokens"], _input["token_price"]
                    )
                )

        return "commons", commons

    @staticmethod
    def su_update_participants_tokens(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        decisions = _input["participant_decisions"]
        final_token_distribution = _input["final_token_distribution"]
        tokens = _input["tokens"]

        for participant_idx, decision in decisions.items():
            network.nodes[participant_idx]["item"].increase_holdings(
                final_token_distribution[participant_idx] * tokens
            )

        return "network", network


class ParticipantSellsTokens:
    @staticmethod
    def p_decide_to_sell_tokens_bulk(params, step, sL, s, **kwargs):
        network = s["network"]
        commons = s["commons"]
        participants = get_participants(network)
        ans = {}
        total_tokens = 0
        for i, participant in participants:
            # If a participant decides to buy, it will be specified in units of DAI.
            # If a participant decides to sell, it will be specified in units of tokens.
            x = participant.sell()
            if x > 0:
                total_tokens += x
                ans[i] = x

        # Now that we have the sum of tokens, ask the Commons object how many
        # DAI would be redeemed as a result. This will be inaccurate due
        # to slippage, and we need the result of this policy to be final to
        # avoid chaining 2 state update functions, so we run the operation on a
        # throwaway copy of Commons
        if total_tokens == 0:
            if params.get("debug"):
                print(
                    "ParticipantSellsTokens: No Participants sold tokens in timestep {}".format(
                        step
                    )
                )
            return {
                "participant_decisions": ans,
                "total_tokens": 0,
                "dai_returned": 0,
                "realized_price": 0,
            }
        else:
            commons2 = copy.copy(commons)
            dai_returned, realized_price = commons2.burn(total_tokens)

            final_dai_distribution = {}
            for i in ans:
                final_dai_distribution[i] = ans[i] / total_tokens

            if params.get("debug"):
                print(
                    "ParticipantSellsTokens: These Participants have decided to sell this many  tokens: {}".format(
                        ans
                    )
                )
                print(
                    "ParticipantSellsTokens: A total of {} tokens will be burned. {} DAI should be returned as a result, at a price of {} DAI/token".format(
                        total_tokens, dai_returned, realized_price
                    )
                )
            return {
                "participant_decisions": ans,
                "total_tokens": total_tokens,
                "dai_returned": dai_returned,
                "realized_price": realized_price,
            }

    @staticmethod
    def su_burn_participants_tokens(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]

        if _input["total_tokens"] > 0:
            dai_returned, realized_price = commons.burn(_input["total_tokens"])
            if (
                _input["dai_returned"] != dai_returned
                or _input["realized_price"] != realized_price
            ):
                raise Exception(
                    "ParticipantSellsTokens: {} DAI was returned at a price of {} (expected: {} with price {})".format(
                        dai_returned,
                        realized_price,
                        _input["dai_returned"],
                        _input["realized_price"],
                    )
                )

        return "commons", commons

    @staticmethod
    def su_update_participants_tokens(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        decisions = _input["participant_decisions"]
        dai_returned = _input["dai_returned"]

        for participant_idx, decision in decisions.items():
            network.nodes[participant_idx]["item"].spend(decision)

        return "network", network


class ParticipantExits:
    @staticmethod
    def p_participant_decides_if_he_wants_to_exit(params, step, sL, s, **kwargs):
        network = s["network"]
        participants = get_participants(network)
        defectors = {}
        for i, participant in participants:
            if participant.wants_to_exit():
                defectors[i] = {
                    "sentiment": participant.sentiment,
                    "holdings": participant.holdings.total,
                }

        if params.get("debug"):
            print(
                "ParticipantExits: Participants {} (2nd number is their sentiment) want to exit".format(
                    defectors
                )
            )
        return {"defectors": defectors}

    @staticmethod
    def su_remove_participants_from_network(params, step, sL, s, _input, **kwargs):
        network = s["network"]
        defectors = _input["defectors"]

        for i, _ in defectors.items():
            network.remove_node(i)

        return "network", network

    @staticmethod
    def su_burn_exiters_tokens(params, step, sL, s, _input, **kwargs):
        commons = s["commons"]
        network = s["network"]
        defectors = _input["defectors"]

        burnt_token_results = {}
        for i, v in defectors.items():
            dai_returned, realized_price = commons.burn(v["holdings"])
            burnt_token_results[i] = {
                "dai_returned": dai_returned,
                "realized_price": realized_price,
            }

        return "commons", commons

    @staticmethod
    def su_update_sentiment_when_proposal_becomes_active(
        params, step, sL, s, _input, **kwargs
    ):
        network = s["network"]
        policy_output_passthru = s["policy_output"]

        report = {}
        for idx in policy_output_passthru["proposal_idxs_with_enough_conviction"]:
            for participant_idx, proposal_idx, _ in find_in_edges_of_type_for_proposal(
                network, idx, "support"
            ):
                edge = network.edges[participant_idx, proposal_idx]
                if edge["support"].is_author:
                    sentiment_old = network.nodes[participant_idx]["item"].sentiment
                    sentiment_new = (
                        sentiment_old + config.sentiment_bonus_proposal_becomes_active
                    )
                    sentiment_new = 1 if sentiment_new > 1 else sentiment_new
                    network.nodes[participant_idx]["item"].sentiment = sentiment_new

                    report[participant_idx] = {
                        "proposal_idx": proposal_idx,
                        "sentiment_old": sentiment_old,
                        "sentiment_new": sentiment_new,
                    }

        if params.get("debug"):
            for i in report:
                print(
                    "ParticipantExits: Participant {} changed his sentiment from {} to {} because Proposal {} became active".format(
                        i,
                        report[i]["sentiment_old"],
                        report[i]["sentiment_new"],
                        report[i]["proposal_idx"],
                    )
                )
        return "network", network

    @staticmethod
    def su_update_sentiment_when_proposal_becomes_failed_or_completed(
        params, step, sL, s, _input, **kwargs
    ):
        network = s["network"]
        policy_output_passthru = s["policy_output"]

        proposal_status_delta = {
            "failed": config.sentiment_bonus_proposal_becomes_failed,
            "succeeded": config.sentiment_bonus_proposal_becomes_completed,
        }
        report = {}
        for status, delta in proposal_status_delta.items():
            for idx in policy_output_passthru[status]:
                for (
                    participant_idx,
                    proposal_idx,
                    _,
                ) in find_in_edges_of_type_for_proposal(network, idx, "support"):
                    edge = network.edges[participant_idx, proposal_idx]
                    # Update the participant sentiment if he/she is the proposal creator
                    # or if participant has staked on the proposal (tokens > 0)
                    if edge["support"].is_author or edge["support"].tokens > 0:
                        sentiment_old = network.nodes[participant_idx]["item"].sentiment
                        sentiment_new = sentiment_old + (
                            edge["support"].affinity * delta
                        )
                        sentiment_new = np.clip(sentiment_new, a_min=0.0, a_max=1.0)
                        network.nodes[participant_idx]["item"].sentiment = sentiment_new

                        report[participant_idx] = {
                            "proposal_idx": proposal_idx,
                            "sentiment_old": sentiment_old,
                            "sentiment_new": sentiment_new,
                            "status": status,
                        }

        if params.get("debug"):
            for i in report:
                print(
                    "ParticipantExits: Participant {} changed his sentiment from {} to {} because Proposal {} became {}".format(
                        i,
                        report[i]["sentiment_old"],
                        report[i]["sentiment_new"],
                        report[i]["proposal_idx"],
                        report[i]["status"],
                    )
                )
        return "network", network


class ParticipantSentiment:
    @staticmethod
    def su_update_sentiment_decay(params, step, sL, s, _input, **kwargs):
        network = s["network"]

        participants = get_participants(network)
        for participant_idx, participant in participants:
            sentiment_old = network.nodes[participant_idx]["item"].sentiment
            sentiment_new = sentiment_old - config.sentiment_decay
            sentiment_new = 0 if sentiment_new < 0 else sentiment_new
            network.nodes[participant_idx]["item"].sentiment = sentiment_new

        return "network", network


end_all(globals())



import numpy as np
from aiondao import config
from aiondao import utils as utilz
from typing import Optional
from aiondao.agents.commons import TokenBatch
from aiondao.interfaces.policy import IPolicyTransform
from aiondao.libs.states import SimState
from aiondao._imports import *
from aiondao.libs.conviction import trigger_threshold
from aiondao.schemas import InvestmentInput

from aiondao.utils.networking import (
    add_participant,
    add_proposal,
    calc_median_affinity,
    calc_total_conviction,
    calc_total_funds_requested,
    find_in_edges_of_type_for_proposal,
    get_edges_by_type,
    get_participants,
    get_proposals,
    get_proposals_by_participant_and_status,
)
from loguru import logger
from aiondao.utils.mixins import StateAccessors
from aiondao.agents.members import Participant, Proposal, ProposalStatus

from aiondao.utils.networking import (
    calc_median_affinity,
    calc_total_funds_requested,
    get_participants,
)
from loguru import logger


class ExamplePolicy(IPolicyTransform):
    def execute(self, state: SimState) -> Optional[SimState]:
        state.policy_log.append({"message": "ExamplePolicy"})
        return state


class SyncState(IPolicyTransform):
    def execute(self, state: SimState) -> Optional[SimState]:
        state.collateral_pool = state.commons._collateral_pool
        state.token_supply = state.commons.token_supply
        state.token_price = state.commons.token_price()
        state.funding_pool = state.commons.funding_pool
        return state


class IStatefulPolicyBlock(StateAccessors, IPolicyTransform, abc.ABC):
    @abc.abstractmethod
    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        # Run the steps here independently and let the state self register
        return self.state

    def execute(self, state: SimState) -> Optional[SimState]:
        self.register_state(state)
        self.transform_steps()
        return self.state


class CreateParticipant(IStatefulPolicyBlock):
    def check_for_and_add_random_entrant(self):
        commons = self.state.commons
        fns = self.state.to_fns()
        sentiment = self.state.sentiment
        timestep = self.state.timestep

        speculation_days = self.state.confs.speculation_days
        multiplier_new_participants = self.state.confs.multiplier_new_participants
        investments = []
        if timestep < speculation_days:
            # If in speculation period, the arrival rate is higher
            arrival_rate = 0.5 + 0.5 * sentiment
            multiplier = multiplier_new_participants
            max_new_participants = config.max_new_participants * multiplier
        else:
            arrival_rate = (1 + sentiment) / config.arrival_rate_denominator
            max_new_participants = config.max_new_participants

        for _ in range(int(max_new_participants)):
            # logger.info((_, max_new_participants))
            if not self.prob_fn(arrival_rate):
                continue
            # Here we randomly generate each participant's post-Hatch
            # investment, in DAI/USD.
            #
            # expon.rvs() arguments:
            #
            # loc is the minimum number, so if loc=100, there will be no
            # investments < 100
            #
            # scale is the standard deviation, so if scale=2, investments will
            # be around 0-12 DAI or even 15, if scale=100, the investments will be
            # around 0-600 DAI.
            participant_investment = fns.exp_fn(
                loc=config.investment_new_participant_min,
                scale=config.investment_new_participant_stdev,
            )

            # Get token conversion
            tokens = commons.dai_to_tokens(participant_investment)

            invest = InvestmentInput(investment=participant_investment, tokens=tokens)
            investments.append(invest)
            if not invest.is_valid():
                continue

            """
                network, i = add_participant(network, Participant(TokenBatch(
                    0, ans["new_participant_tokens"]), probability_func,
                    random_number_func), exponential_func, random_number_func)

                if params.get("debug"):
                    print("GenerateNewParticipant: A new Participant {} invested {}DAI for {} tokens".format(
                        i, ans['new_participant_investment'], ans['new_participant_tokens']))
            
            """

            network, idx = add_participant(
                self.network,
                Participant(
                    holdings=TokenBatch(vesting=0, nonvesting=invest.tokens),
                    prob_fn=self.prob_fn,  # type: ignore
                    random_number_func=self.rand_fn,  # type: ignore
                ),
                self.exp_fn,
                self.rand_fn,
            )
            self.state.network = network

            self.commons.deposit(invest.investment)

    def update_token_batch_ages(self):
        network = self.state.network
        participants = get_participants(network)
        for _, participant in participants:
            participant = cast(Participant, participant)

            participant.update_token_batch_age()
        self.state.network = network

    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        self.check_for_and_add_random_entrant()
        self.update_token_batch_ages()
        return super().transform_steps(state)


class CreateProposal(IStatefulPolicyBlock):
    def random_generate_proposal(self) -> Optional[SimState]:
        participants_dict = self.gx.parts.all_dict()
        # participants_dict = dict(participants)
        selected_idx = self.state.confs.choice_func(list(participants_dict.keys()))
        participant = participants_dict[selected_idx]

        wants_to_create_proposal = participant.create_proposal(
            calc_total_funds_requested(self.network),
            calc_median_affinity(self.network),
            self.state.funding_pool,
        )
        return wants_to_create_proposal, selected_idx

    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        created_proposal, proposed_by = self.random_generate_proposal()
        if not created_proposal:
            return state
        rescale = self.funding_pool * config.scale_factor
        requested_funds = self.gamma_func(
            config.funds_requested_alpha,
            loc=config.funds_requested_min,
            scale=rescale,
        )
        proposal = Proposal(
            funds_requested=requested_funds,
            trigger=trigger_threshold(
                requested_funds,
                self.funding_pool,
                self.token_supply,
                self.cp.max_proposal_request,
            ),
        )
        network, _ = add_proposal(self.network, proposal, proposed_by, self.rand_fn)
        self.state.network = network


class CreateFunding(IStatefulPolicyBlock):
    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        exits = [
            self.exp_fn(
                loc=config.speculator_position_size_min,
                scale=config.speculator_position_size_stdev,
            )
            for i in range(config.speculators)
        ]
        funding = sum(exits) * self.commons.exit_tribute
        self.commons._funding_pool += funding
        return self.state


class ProcessActiveProposals(IStatefulPolicyBlock):
    def determine_proposal_outcomes(self):
        probability_func = self.prob_fn

        active_proposals = get_proposals(self.network, status=ProposalStatus.ACTIVE)
        failed_procedures = []
        successul_procedures = []
        for idx, proposal in active_proposals:
            r_failure = 1 / (
                config.base_failure_rate + np.log(proposal.funds_requested)
            )
            r_success = 1 / (
                config.base_success_rate + np.log(proposal.funds_requested)
            )
            if probability_func(r_failure):
                failed_procedures.append(idx)
            elif probability_func(r_success):
                successul_procedures.append(idx)

        return failed_procedures, successul_procedures

    def update_proposal_statuses(self):
        failed, success = self.determine_proposal_outcomes()
        self.state.policy_output["failed"] = failed
        self.state.policy_output["success"] = success
        for idx in failed:
            self.network.nodes[idx]["item"].status = ProposalStatus.FAILED

        for idx in success:
            self.network.nodes[idx]["item"].status = ProposalStatus.COMPLETED

        return self.state

    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        self.update_proposal_statuses()
        return self.state


class ParticipantVoting(IStatefulPolicyBlock):
    # Get voite participantion
    def get_participant_votes(self):
        """
        p_participant_votes_on_proposal_according_to_affinity
        # We generate the probability of voting on a proposal according to the affinity to the proposal.
        This policy collects data from the DiGraph to tell the Participant class
        which candidate proposals it supports, and
        Participant.vote_on_candidate_proposals() will decide which proposals it
        will take action on.

        Then, Participant.stake_across_all_supported_proposals() will tell us
        how much it will stake on each of them.
        """
        participants = get_participants(self.network)

        participants_stakes = {}
        for participant_idx, participant in participants:
            proposal_idx_affinity = {}  # {4: 0.9, 5: 0.9}
            candidate_proposals = get_proposals_by_participant_and_status(
                self.network,
                participant_idx=participant_idx,
                status_filter=[ProposalStatus.CANDIDATE],
            )
            for proposal_idx, proposal in candidate_proposals.items():
                proposal_idx_affinity[proposal_idx] = proposal["support"].affinity
            proposals_that_participant_cares_enough_to_vote_on = (
                participant.vote_on_candidate_proposals(proposal_idx_affinity)
            )

            stake_across_all_supported_proposals_input = []
            for (
                proposal_idx,
                affinity,
            ) in proposals_that_participant_cares_enough_to_vote_on.items():
                stake_across_all_supported_proposals_input.append(
                    (affinity, proposal_idx)
                )
            stakes = participant.stake_across_all_supported_proposals(
                stake_across_all_supported_proposals_input
            )

            participants_stakes[participant_idx] = stakes

            # if params.get("debug"):
            # if proposals_that_participant_cares_enough_to_vote_on:
            #     logger.warning(
            #         f"ParticipantVoting: Participant {participant_idx} was given Proposals with corresponding affinities {proposal_idx_affinity} and he decided to vote on {proposals_that_participant_cares_enough_to_vote_on}, distributing his tokens thusly {stakes}"
            #     )
        return participants_stakes

    def update_participants_votes(self):
        """
        Simply update the support edges with the new amount of tokens the
        Participant has staked on the Proposal. Leave the conviction calculation
        to another state update function.
        """
        network = self.network

        for participant_idx, v in self.get_participant_votes().items():

            for proposal_idx, tokens_staked in v.items():
                # I assume there is no longer a need to recalculate a la
                # https://github.com/randomshinichi/conviction/blob/9d1bc9513475dc30d33e3232b385234d0295d361/conviction_system_logic3.py#L510
                # because the tokens were already distributed proportionally to
                # the affinities in
                # p_participant_votes_on_proposal_according_to_affinity()
                # Also, do not recalculate conviction here. Leave that to ProposalFunding.su_calculate_conviction()
                start_support = network[participant_idx][proposal_idx]["support"]
                start_support.tokens = tokens_staked
                network[participant_idx][proposal_idx]["support"] = start_support

        self.state.network = network
        return self.state

    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        self.get_participant_votes()
        return self.update_participants_votes()


class ProposalFunding(IStatefulPolicyBlock):
    """Determine if there are any proposals that should go on to get fundint.

    If there are, then determine how much funding should be requested.
    """

    def find_approved_proposals(self):
        """
        This policy simply goes through the Proposals to see if their thresholds
        are smaller than their gathered conviction. If the conviction is higher, then pass the proposals to be activated.
        """

        proposals_w_enough_conviction = []
        proposals = get_proposals(self.network, status=ProposalStatus.CANDIDATE)
        for idx, proposal in proposals:
            total_conviction = calc_total_conviction(self.network, idx)
            proposal.conviction = total_conviction
            is_above_threshold = proposal.has_enough_conviction(
                self.funding_pool, self.token_supply, self.capt.max_proposal_request
            )

            if is_above_threshold:
                proposals_w_enough_conviction.append(idx)

        return proposals_w_enough_conviction

    def process_activated_proposals(self):
        active_proposals = self.find_approved_proposals()
        self.state.conviction_proposals = active_proposals
        for idx in active_proposals:
            self.network.nodes[idx]["item"].status = ProposalStatus.ACTIVE
            funds = self.network.nodes[idx]["item"].funds_requested
            self.commons.spend(funds)
        return self.state

    def update_age_and_conviction_thresholds(self):

        proposals = get_proposals(self.network, status=ProposalStatus.CANDIDATE)
        for _, proposal in proposals:
            proposal.update_age()
            proposal.update_threshold(
                self.funding_pool,
                self.token_supply,
                max_proposal_request=self.capt.max_proposal_request,
            )

        return self.state

    def recalculate_convictions(self):
        """
        Actually calculates the conviction. This function should only run ONCE
        per timestep/iteration!
        """
        alpha = self.cp.days_to_80p_of_max_voting_weight

        support_edges = get_edges_by_type(self.network, "support")
        for i, j in support_edges:
            proposal_status = self.network.nodes[j]["item"].status
            if proposal_status == ProposalStatus.CANDIDATE:
                edge = self.network.edges[i, j]
                prior_conviction = edge["support"].conviction
                current_tokens = edge["support"].tokens
                local_support = edge["support"]
                local_support.conviction = current_tokens + alpha * prior_conviction
                edge["support"] = local_support
                self.network.edges[i, j]["support"] = edge["support"]

        return self.state

    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        self.process_activated_proposals()
        self.update_age_and_conviction_thresholds()
        self.recalculate_convictions()

        return self.state


class ParticipantSentiment(IStatefulPolicyBlock):
    def transform_steps(self, state: Optional[SimState] = None) -> Optional[SimState]:
        participants = get_participants(self.state.network)
        for participant_idx, participant in participants:
            sentiment_old = self.state.network.nodes[participant_idx]["item"].sentiment
            sentiment_new = sentiment_old - config.sentiment_decay
            sentiment_new = 0 if sentiment_new < 0 else sentiment_new
            self.state.network.nodes[participant_idx]["item"].sentiment = sentiment_new

        return super().transform_steps(self.state)

    # def execute(self, state: SimState) -> Optional[SimState]:
    #     logger.debug("Create new participant")
    #     state.collateral_pool = state.commons._collateral_pool
    #     state.token_supply = state.commons.token_supply
    #     state.token_price = state.commons.token_price()
    #     state.funding_pool = state.commons.funding_pool
    #     return state



from typing import Any
from pydantic import Extra
from pydantic.dataclasses import dataclass
from aiondao._imports import *


# value function for a given state (reserve,supply)
def invariant(reserve: float, supply: float, kappa: float):
    return (supply**kappa) / reserve


# value function for a given state (R,S)
def invariant(reserve: float, supply: float, power_invariant: float):

    return (supply**power_invariant) / reserve


# given a value function (parameterized by kappa)
# and an invariant coefficient invariant
# return Supply supply as a function of reserve


def supply(reserve: float, kappa: float, invariant: float):
    return (invariant * reserve) ** (1 / kappa)


# given a value function (parameterized by kappa)
# and an invariant coeficient invariant
# return a spot price P as a function of reserve


def spot_price(reserve: float, kappa: float, invariant: float):
    return kappa * reserve ** ((kappa - 1) / kappa) / invariant ** (1 / kappa)


# for a given state (reserve,supply)
# given a value function (parameterized by kappa)
# and an invariant coeficient invariant
# deposit d_reserve to Mint d_supply
# with realized price d_reserve/d_supply


def mint(
    d_reserve: float, reserve: float, supply: float, kappa: float, invariant: float
):
    d_supply = (invariant * (reserve + d_reserve)) ** (1 / kappa) - supply
    realized_price = d_reserve / d_supply
    return d_supply, realized_price


# for a given state (reserve,supply)
# given a value function (parameterized by kappa)
# and an invariant coeficient invariant
# burn d_supply to Withdraw d_reserve
# with realized price d_reserve/d_supply


def withdraw(
    d_supply: float, reserve: float, supply: float, kappa: float, invariant: float
):
    d_reserve = reserve - ((supply - d_supply) ** kappa) / invariant
    realized_price = d_reserve / d_supply
    return d_reserve, realized_price


class AugmentedBondingCurve(BaseModel):
    """Create a stateless bonding curve.

    reserve_initial (DAI)
    token_supply_initial (DAI)
    kappa (the exponent part of the curve, default is 2)
    """

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow
        validate_assignment = True

    reserve: float
    supply: float
    kappa: float
    invariant: Optional[float] = None
    _has_init: bool = PrivateAttr(False)

    @root_validator
    def calc_invariant(cls, values):
        reserve = values.get("reserve")
        supply = values.get("supply")
        kappa = values.get("kappa")
        values["invariant"] = invariant(
            reserve=reserve, supply=supply, power_invariant=kappa
        )
        return values

    # def __post_init__(self):
    #     self.invariant = invariant(self.reserve, self.supply, self.kappa)

    # def __repr__(self):
    #     return su

    def deposit(self, dai: float, current_reserve: float, current_token_supply: float):
        # Returns number of new tokens minted, and their realized price
        tokens, realized_price = mint(
            dai, current_reserve, current_token_supply, self.kappa, self.invariant
        )
        return tokens, realized_price

    def burn(
        self,
        tokens_millions: float,
        current_reserve: float,
        current_token_supply: float,
    ):
        # Returns number of DAI that will be returned (excluding exit tribute) when the user burns their tokens, with their realized price
        dai, realized_price = withdraw(
            tokens_millions,
            current_reserve,
            current_token_supply,
            self.kappa,
            cast(float, self.invariant),
        )
        return dai, realized_price

    def get_token_price(self, current_reserve: float):
        return spot_price(current_reserve, self.kappa, self.invariant)

    def get_token_supply(self, current_reserve: float):
        return supply(current_reserve, self.kappa, self.invariant)

    def update(self, reserve: float, supply: float):
        self.init_reserve = reserve
        self.supply = supply



import numpy as np
from aiondao import config
from aiondao._imports import *


def trigger_threshold(
    funds_requested, funding_pool, token_supply, max_proposal_request
):
    """
    funds_requested: funds requested by the proposal
    funding_pool: the current size of the funding pool
    token_supply: current token_supply
    max_proposal_request: maximum fraction of the funding pool that a proposal can ever request
    """
    rho = config.rho_multiplier * max_proposal_request**config.rho_power

    fraction = funds_requested / funding_pool
    if fraction < max_proposal_request:
        return rho * token_supply / (max_proposal_request - fraction) ** 2
    else:
        return np.inf



"""
The following code was adapted from the Bank Reserves model included in Netlogo
Model information can be found at: http://ccl.northwestern.edu/netlogo/models/BankReserves
Accessed on: November 2, 2017
Author of NetLogo code:
    Wilensky, U. (1998). NetLogo Bank Reserves model.
    http://ccl.northwestern.edu/netlogo/models/BankReserves.
    Center for Connected Learning and Computer-Based Modeling,
    Northwestern University, Evanston, IL.
"""

# isort: off
from typing import Iterable, cast
from aiondao._imports import *

# isort: on
import random

import numpy as np
from devtools import debug

from aiondao.agents.members import Participant

from aiondao.agents.commons import (
    CommonAgent,
    create_token_batches,
)

from loguru import logger
from mesa import Agent, Model
from mesa.time import RandomActivation, BaseScheduler
from aiondao.utils.networking import bootstrap_network, get_participants
from aiondao.libs.states import CommonsConfig, SimState
from aiondao.interfaces.models import IEconomy
from aiondao.libs.monitors import (
    get_num_rich_agents,
    get_num_poor_agents,
    get_num_mid_agents,
    get_total_savings,
    get_total_wallets,
    get_total_money,
    get_total_loans,
)
from aiondao.libs.bootloader import BootstrapSimulation
from aiondao.policies.transformer import (
    SyncState,
    CreateParticipant,
    CreateProposal,
    CreateFunding,
    ProcessActiveProposals,
    ParticipantVoting,
    ProposalFunding,
    ParticipantSentiment,
)
from aiondao.libs.pipes import Sequential
from aiondao.libs.graph import GraphORM

from aiondao.libs.states import SimState, StepSpace


class _AionEconomy(IEconomy, Model):

    schedule: "BaseScheduler"
    state: "SimState"

    @property
    def gfx(self) -> GraphORM:
        return GraphORM(net=self.state.network)

    @property
    def booter(self) -> BootstrapSimulation:
        if not hasattr(self, "_bootstrapper"):
            self._bootstrapper = BootstrapSimulation()
        return self._bootstrapper

    def start_step(self):
        pass
        # return super().start_step()

    def end_step(self):
        pass


class AionEconomy(_AionEconomy):
    """Transferred DAO"""

    """init parameters "init_people", "rich_threshold", and "reserve_percent"
       are all UserSettableParameters"""
    __slots__ = ["state", "params", "schedule", "policies", "count"]

    def __init__(self, params: CommonsConfig):
        super(AionEconomy, self).__init__()
        self.required()
        self.state: SimState = self.bootstrap(params)

        self.commons: CommonAgent = self.state.commons
        self.schedule.add(self.commons)

    def required(self):
        # self.count = 0
        self.policies = Sequential(
            [
                SyncState(),
                CreateParticipant(),
                CreateProposal(),
                CreateFunding(),
                # Should update the sentiment here as well
                SyncState(),
                ProcessActiveProposals(),
                ParticipantVoting(),
                ProposalFunding(),
                # Add more sentiment here
                ParticipantSentiment(),
                SyncState(),
            ]
        )
        self.post_policies = Sequential(
            [
                SyncState(),
                ProcessActiveProposals(),
                ParticipantVoting(),
                ProposalFunding(),
                # Add more sentiment here
                ParticipantSentiment(),
                SyncState(),
            ]
        )

        self.pre_policies = Sequential(
            [
                SyncState(),
                CreateParticipant(),
                CreateProposal(),
                CreateFunding(),
                # Should update the sentiment here as well
            ]
        )

    @property
    def participants(self) -> Iterable[Participant]:

        return []

    def bootstrap(self, params: CommonsConfig):
        self.run_time = params.run_time
        self.params = params
        self.schedule: BaseScheduler = RandomActivation(self)
        """Sets up the simulation variables, states, and agents."""

        return self.booter.bootload(params, self)

    def get_flags(self):
        return StepSpace()

    def start_step(self):
        pass
        # return super().start_step()

    def end_step(self):
        pass
        # return super().end_step()

    def step(self):
        # tell all the agents in the model to run their step function
        # Just resetiing a variable that needsd to be reset before processing
        # Collecting code
        # Collect data for the simulation using mesa's data collection code.
        # https://mesa.readthedocs.io/en/stable/tutorials/intro_tutorial.html#collecting-data

        self.state.conviction_proposals = []
        self.state.policy_log = {}
        # logger.debug(self.state)
        self.flags = self.get_flags()
        self.state = self.pre_policies.transform(self.state)

        self.schedule.step()
        self.state = self.post_policies.transform(self.state)
        # Here we update the state of the model
        self.state.timestep = self.schedule.time

    def run_model(self):
        for i in range(self.run_time):
            self.step()


def main():
    pass


if __name__ == "__main__":
    main()



from typing import List, Optional, Tuple, Union
from aiondao._imports import *
from aiondao.agents.members import Participant, Proposal, ProposalStatus
from aiondao.utils.networking import (
    add_proposal,
    get_participants,
    get_proposals,
    calc_total_funds_requested,
    calc_total_funds_requested,
    add_participant,
    calc_median_affinity,
    calc_total_conviction,
    calc_total_affinity,
    calc_avg_sentiment,
)


import networkx as nx
import toolz as tlz


class _BaseModel(RelaxedModel):
    class Config:
        arbitrary_types_allowed = True


# Proposals and Participants queries. To be accessed by `GraphORM` class decorators parts and props`
class ProposalsQuery(BaseModel):
    net: nx.DiGraph

    def all(self, status: Optional[ProposalStatus] = None) -> List[Participant]:
        return list(dict(get_proposals(self.net, status)).values())

    def add(
        self, proposal: Proposal, proposed_by: int, random_number_func
    ) -> Tuple[nx.DiGraph, int]:
        return add_proposal(self.net, proposal, proposed_by, random_number_func)

    def by_parts_and_type(self, participant: Participant, type: str) -> List[Proposal]:
        return []

    def by_parts_and_stats(self, participant: Participant, type: str) -> List[Proposal]:
        return []

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow


class ParticipantsQuery(_BaseModel):
    net: nx.DiGraph

    def all(self) -> List[Proposal]:
        return list(dict(get_participants(self.net)).values())

    def all_dict(self) -> Dict[int, Proposal]:
        return dict(get_participants(self.net))

    def add(
        self, participant: Participant, exponential_func, random_number_func
    ) -> Tuple[nx.DiGraph, int]:
        return add_participant(
            self.net, participant, exponential_func, random_number_func
        )

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow


# A dataclass for graph orm with property functions for 'participants' and 'proposals'
class GraphORM(BaseModel):

    net: nx.DiGraph

    @property
    def parts(self) -> ParticipantsQuery:
        return ParticipantsQuery(net=self.net)

    @property
    def props(self) -> ProposalsQuery:
        return ProposalsQuery(net=self.net)

    @property
    def total_funds_requested(self) -> float:
        """The total_funds_requested property."""
        return calc_total_funds_requested(self.net)

    @property
    def median_affinity(self) -> float:
        """The total_funds_requested property."""
        return calc_median_affinity(self.net)

    @property
    def total_conviction(self) -> float:
        """The total_funds_requested property."""
        return calc_total_conviction(self.net, 0)

    @property
    def total_affinity(self) -> float:
        """The total_funds_requested property."""
        return calc_total_affinity(self.net)

    @property
    def avg_sentiment(self) -> float:
        """The total_funds_requested property."""
        return calc_avg_sentiment(self.net)

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow



from mesa.agent import Agent

# isort: off
from typing import cast

# isort: on


# from aiondao.agents.dummy import DummyAgent

from loguru import logger
from mesa import Agent
from aiondao.libs.states import SimState

from loguru import logger

ilog = logger.info


class DummyAgent(Agent):
    """An agent with fixed initial wealth."""

    def __init__(self, unique_id, model):
        super().__init__(unique_id, model)
        self.wealth = 1

        logger.debug(f"Starting Dummy Model: {self.unique_id}")

    def step(self):
        self.model.commons = cast(SimState, self.model.commons)
        self.model.commons.unique_id += 1

        return super().step()



from aiondao._imports import *
from aiondao.economy import AionEconomy
from aiondao.libs.states import CommonsConfig


def run_dev():
    logger.info("----- Starting AionDAO DEV -----")
    conf = CommonsConfig()
    economy = AionEconomy(conf)

    # model = economy
    for i in range(20):
        economy.step()


if __name__ == "__main__":
    run_dev()



from aiondao._imports import *
from typing import Any, Dict, List, Optional
from pydantic import BaseModel


class InvestmentInput(RelaxedModel):
    investment: float
    tokens: float

    def is_valid(self):
        return True if self.investment > 0 and self.tokens > 0 else False


class ParticipantPurchase(SyncModel):
    participant_decisions: List[InvestmentInput]
    total_dai: float
    tokens: float
    token_price: float
    final_token_distribution: Dict[str, Any] = {}
    dai_returned: float = 0
    realized_price: float = 0


class PurchaseEvent(BaseModel):
    participant_decisions: List[InvestmentInput]
    total_dai: float
    tokens: float
    token_price: float
    final_token_distribution: Dict[str, Any] = {}
    dai_returned: float = 0
    realized_price: float = 0


class Defector(RelaxedModel):
    sentiment: float = 0.75
    holdings: float


class PolicyAggregate(RelaxedModel):
    invests: List[InvestmentInput] = []
    purchase: Optional[ParticipantPurchase] = None
    decision: Optional[ParticipantPurchase] = None
    total_tokens: float = 0
    dai_returned: float = 0
    realized_price: float = 0
    defectors: Dict[str, Defector] = {}



from aiondao.libs.states import SimState
from aiondao.libs.graph import GraphORM


class StateAccessors:
    def __init__(self) -> None:
        self.__state = None

    # Properties to access network, functinf pool, and graphorm queries
    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, state: SimState):
        self.__state = state

    @property
    def network(self):
        return self.__state.network

    @property
    def capt(self):
        return self.__state.confs

    @property
    def cp(self):
        return self.__state.confs

    @property
    def commons(self):
        return self.__state.commons

    @property
    def gx(self):
        return GraphORM(net=self.__state.network)

    @property
    def funding_pool(self):
        return self.__state.funding_pool

    @property
    def token_supply(self):
        return self.__state.commons.token_supply

    @property
    def gamma_func(self):
        return self.__state.confs.gamma_func

    @property
    def participants(self):
        return self.gx.parts.all()

    @property
    def proposals(self):
        return self.gx.props.all()

    @property
    def exp_fn(self):
        return self.__state.confs.exponential_func

    @property
    def prob_fn(self):
        return self.__state.confs.probability_func

    @property
    def rand_fn(self):
        return self.__state.confs.random_number_func

    def register_state(self, state: SimState):
        self.state = state



__version__ = "0.1.0"

# from typing import Any, Callable, Type, Union, TYPE_CHECKING, TypeVar, cast
# from pydantic import BaseConfig, Extra
# from pydantic.dataclasses import dataclass as _dataclass

# if TYPE_CHECKING:
#     from typing import Dataclass


# from functools import partial


# class BroadConfig(BaseConfig):
#     arbitrary_types_allowed = True
#     underscore_attrs_are_private = True


# class cfgall(BroadConfig):

#     extra = Extra.allow


# class cfgign(BroadConfig):
#     extra = Extra.ignore


# class autovalid(BaseConfig):
#     """Automatically validate assigned values."""

#     extra = Extra.allow
#     validate_assignment = True


# class iautovalid(BaseConfig):
#     """Automatically validate assigned values. with ignore extra."""

#     extra = Extra.ignore
#     validate_assignment = True


# dataclass = partial(_dataclass, config=autovalid)


# dataclass = cast(_dataclass, dataclass)
# # dataclass = dataclass(config=autovalid)
# # idataclass = dataclass(config=autovalid)



from pydantic import BaseConfig, Extra

# Participant Sentiment Parameters
sentiment_decay: float = 0.005
sentiment_sensitivity: float = 0.75
candidate_proposals_cutoff: float = 0.75
delta_holdings_scale: float = 70000
sentiment_bonus_proposal_becomes_active: float = 0.5
sentiment_bonus_proposal_becomes_completed: float = 0.3
sentiment_bonus_proposal_becomes_failed: float = -0.1
engagement_rate_multiplier_buy: float = 0.6
engagement_rate_multiplier_sell: float = 0.6
engagement_rate_multiplier_exit: float = 0.3
sentiment_sensitivity_exit: float = 0.5

# Participant Investment Parameters
investment_new_participant_min: float = 0.0
investment_new_participant_stdev: float = 100

# Speculator Parameters
speculator_position_size_min: float = 200  # DAI
speculator_position_size_stdev: float = 200
speculators: int = 5

# Proposal Parameters
min_age_days: int = 2
scale_factor: float = 0.01
base_failure_rate: float = 0.15
base_success_rate: float = 0.30
funds_requested_alpha: int = 3
funds_requested_min: float = 0.001

# Trigger threshold Parameters
rho_multiplier: float = 0.5
rho_power: float = 2

# Vesting curve Parameters
vesting_curve_halflife: float = 0.5
log_base05_of_02: float = 2.321928094887362

# GenerateNewParticipant Parameters
arrival_rate_denominator: float = 10
max_new_participants: int = 1


class BroadConfig(BaseConfig):
    arbitrary_types_allowed = True


class cfgall(BroadConfig):

    extra = Extra.allow


class cfgign(BroadConfig):
    extra = Extra.ignore


class autovalid(BroadConfig):
    """Automatically validate assigned values."""

    extra = Extra.allow
    validate_assignment = True


class iautovalid(BroadConfig):
    """Automatically validate assigned values. with ignore extra."""

    extra = Extra.ignore
    validate_assignment = True



"""Base blocks for simulation and changes the state of the model."""
import abc
from typing import List, Optional
from aiondao.libs.states import SimState


class ISimBlock(abc.ABC):
    """
    A simulation block is a part of the simulation that is executed in a single
    step. It recieves the state and returns the new state. It contains a single transformation that can get built upon over time.
    """

    @abc.abstractmethod
    def execute(self, state: SimState) -> Optional[SimState]:
        """
        The small units of work.
        """


class ISimTransform(ISimBlock, abc.ABC):
    def transform(self, state_space: SimState) -> SimState:
        updated: SimState = self.execute(state_space)
        assert (
            updated is not None
        ), "SimTransform must return a new state. None is not allowed"
        return updated


class IPipeline(ISimBlock, abc.ABC):
    def __init__(self, state_transforms: List[ISimTransform]) -> None:
        if not state_transforms:
            raise ValueError(
                "PolicySequential requires at least one state transform. Declaring it without using it is not allowed"
            )
        self.state_transforms: List[ISimTransform] = state_transforms

    def transform(self, state_space: SimState) -> SimState:
        for state_transformer in self.state_transforms:
            state_space = state_transformer.transform(state_space)
        return state_space



from typing import List, Optional
from aiondao.libs.states import SimState
from aiondao.interfaces.policy import IPipeline
from aiondao.interfaces.simulate import ISimTransform


class Sequential(IPipeline):
    """Gets the current state anc makes updates to it. Not changing system much compared to the original code so it will happen in batch."""

    def __init__(self, state_transforms: List[ISimTransform]) -> None:
        super().__init__(state_transforms)

    def execute(self, state: SimState) -> Optional[SimState]:
        return super().execute(state)

    def transform(self, state_space: SimState) -> SimState:
        for state_transformer in self.state_transforms:
            state_space = state_transformer.transform(state_space)
        return state_space



"""Global imports to reduce typing."""
# A tacky tactic I picked up to reduce the number of imports to manually do.
# Can use autoflake expand_star_imports to get the specific imports.
__version__ = "0.1.0"

# isort: off
from auto_all import end_all, start_all
from pydantic.main import ModelMetaclass
from pydantic.fields import FieldInfo, ModelField
from pydantic.typing import AnyCallable

start_all(globals())

# isort: on
import abc
import math
from datetime import timedelta
from enum import Enum
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import pandas as pd
from cytoolz.dicttoolz import itemfilter, valfilter, keyfilter
from decorator import decorator
from loguru import logger
from pydantic import (
    BaseConfig,
    BaseModel,
    Extra,
    Field,
    parse_obj_as,
    root_validator,
    PrivateAttr,
    validator,
)
from pydantic.dataclasses import dataclass
from pydantic.generics import GenericModel
from toolz import keyfilter
from pydantic import BaseModel
import networkx as nx

# from toolz import valfilter

ROOT = Path(__file__).parent


CallableAny = Callable[..., Any]
OptCallAny = Optional[CallableAny]
TupleAny = Tuple[str, Any]
StateSP = TypeVar("StateSP")
SysParams = TypeVar("SysParams")
PolicySP = TypeVar("PolicySP")
DictAny = Dict[str, Any]
RandomNumCallable = Callable[..., float]


iprint = logger.info
ilog = logger.info


class StrEnum(str, Enum):
    """Enum with string values."""


class BroadConfig(BaseConfig):
    arbitrary_types_allowed = True
    underscore_attrs_are_private = True


class cfgall(BroadConfig):

    extra = Extra.allow


class cfgign(BroadConfig):
    extra = Extra.ignore


class autovalid(BroadConfig):
    """Automatically validate assigned values."""

    arbitrary_types_allowed = True
    extra = Extra.allow
    validate_assignment = True


class iautovalid(BaseConfig):
    """Automatically validate assigned values. with ignore extra."""

    extra = Extra.ignore
    validate_assignment = True


# import cytoolz
from devtools import debug


class RelaxedModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow
        # validate_assignment = True


_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


@__dataclass_transform__(field_descriptors=(Field, FieldInfo))
class SyncMetadata(ModelMetaclass):
    def __new__(cls, name, bases, namespace, **kwargs):
        """Completely ignore the metadata."""
        return super().__new__(cls, name, bases, namespace, **kwargs)


class SyncModel(BaseModel, metaclass=SyncMetadata):
    __model_maps__: ClassVar[Dict[Union[str, int], Any]] = {}

    if TYPE_CHECKING:
        # populated by the metaclass, defined here to help IDEs only
        __fields__: ClassVar[Dict[str, ModelField]] = {}
        __include_fields__: ClassVar[Optional[Mapping[str, Any]]] = None
        __exclude_fields__: ClassVar[Optional[Mapping[str, Any]]] = None
        __validators__: ClassVar[Dict[str, AnyCallable]] = {}
        __pre_root_validators__: ClassVar[List[AnyCallable]]
        __post_root_validators__: ClassVar[List[Tuple[bool, AnyCallable]]]
        __config__: ClassVar[Type[BaseConfig]] = BaseConfig
        __json_encoder__: ClassVar[Callable[[Any], Any]] = lambda x: x
        __schema_cache__: ClassVar["DictAny"] = {}
        __custom_root_type__: ClassVar[bool] = False

    class Config:
        extra = Extra.allow
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


@__dataclass_transform__(field_descriptors=(Field, FieldInfo))
class BorgModel(SyncModel, metaclass=SyncMetadata):
    __shared_state__: ClassVar[Dict[str, Any]] = {}

    # New class __new__  that inherits the shares state
    def __init__(__pydantic_self__, **data: Any) -> None:
        super().__init__(**data)
        __pydantic_self__.__sync_shared()

    def __setattr__(self, name, value):
        resp = super().__setattr__(name, value)
        self.__sync_shared()
        return resp

    def __getattr__(self, name):
        if name in self.__shared_state__:
            return self.__shared_state__[name]
        try:
            return self.__dict__[name]

        except KeyError as ae:
            raise AttributeError(f"The attribute '{name}' is not defined")

    def __sync_shared(self):
        self.__shared_state__.update(valfilter(lambda x: x is not None, self.__dict__))


@__dataclass_transform__(field_descriptors=(Field, FieldInfo))
class BorgIdModel(SyncModel, metaclass=SyncMetadata):
    # If there's a crazy ID I will
    __shared_state__: ClassVar[Dict[str, Any]] = {}
    crazy_id: Optional[int] = None
    # New class __new__  that inherits the shares state
    def __init__(__pydantic_self__, **data: Any) -> None:
        super().__init__(**data)
        # __pydantic_self__.crazy_id: Optional[int] = None
        if __pydantic_self__.is_crazy():
            __pydantic_self__.__sync_shared()

        # __pydantic_self__.__crazy_id = self.__id
        # debug(__pydantic_self__.__shared_state__)

    def is_crazy(self) -> bool:
        """Detects if there's an id set.

        Returns:
            bool: _description_
        """
        return bool(self.crazy_id)

    # @property
    # def crazy_id(self):
    #     """The crazy_id property."""
    #     return self._crazy_id

    # @crazy_id.setter
    # def crazy_id(self, value: int):
    #     self._crazy_id = value
    #     self.__sync_shared()
    #     debug(self.__shared_state__)

    def is_within(self) -> bool:
        if self.is_crazy() and self.crazy_id in self.__shared_state__:
            return True
        return False

    def __setattr__(self, name, value):
        resp = super().__setattr__(name, value)
        if self.is_crazy():
            self.__sync_shared()

        return resp

    def __getattr__(self, name):
        if (
            self.is_crazy()
            and self.is_within()
            and name in self.__shared_state__[self.crazy_id]
        ):
            return self.__shared_state__.get(self.crazy_id, {}).get(name, None)

        try:
            return self.__dict__[name]

        except KeyError as ae:
            raise AttributeError(f"The attribute '{name}' is not defined")

    def __sync_shared(self):
        idx = self.crazy_id
        # logger.warning(idx)
        removed_none = valfilter(lambda x: (x is not None), self.__dict__)
        removed_id = keyfilter(lambda x: x != "_crazy_id", removed_none)
        curr = self.__shared_state__.get(idx, {})
        curr.update(removed_none)
        # updated_dict_state = self.__shared_state__.get(idx, {}).update(removed_id)
        # debug(curr)
        self.__shared_state__[self.crazy_id] = curr
        # updated_dict_state


# isort: off
end_all(globals())
# isort: on



"""
The following code was adapted from the Bank Reserves model included in Netlogo
Model information can be found at: http://ccl.northwestern.edu/netlogo/models/BankReserves
Accessed on: November 2, 2017
Author of NetLogo code:
    Wilensky, U. (1998). NetLogo Bank Reserves model.
    http://ccl.northwestern.edu/netlogo/models/BankReserves.
    Center for Connected Learning and Computer-Based Modeling,
    Northwestern University, Evanston, IL.
"""

from mesa import Agent
from aiondao.agents.random_walk import RandomWalker


class AionDAO(Agent):
    def __init__(self, unique_id, model, reserve_percent=50):
        # initialize the parent class with required parameters
        super().__init__(unique_id, model)
        # for tracking total value of loans outstanding
        self.bank_loans = 0
        """percent of deposits the bank must keep in reserves - this is a
           UserSettableParameter in server.py"""
        self.reserve_percent = reserve_percent
        # for tracking total value of deposits
        self.deposits = 0
        # total amount of deposits in reserve
        self.reserves = (self.reserve_percent / 100) * self.deposits
        # amount the bank is currently able to loan
        self.bank_to_loan = 0

    """update the bank's reserves and amount it can loan;
       this is called every time a person balances their books
       see below for Person.balance_books()"""

    def bank_balance(self):
        self.reserves = (self.reserve_percent / 100) * self.deposits
        self.bank_to_loan = self.deposits - (self.reserves + self.bank_loans)


# subclass of RandomWalker, which is subclass to Mesa Agent
class Person(RandomWalker):
    def __init__(self, unique_id, pos, model, moore, bank, rich_threshold):
        # init parent class with required parameters
        super().__init__(unique_id, pos, model, moore=moore)
        # the amount each person has in savings
        self.savings = 0
        # total loan amount person has outstanding
        self.loans = 0
        """start everyone off with a random amount in their wallet from 1 to a
           user settable rich threshold amount"""
        self.wallet = self.random.randint(1, rich_threshold + 1)
        # savings minus loans, see balance_books() below
        self.wealth = 0
        # person to trade with, see do_business() below
        self.customer = 0
        # person's bank, set at __init__, all people have the same bank in this model
        self.bank = bank

    def do_business(self):
        """check if person has any savings, any money in wallet, or if the
        bank can loan them any money"""
        if self.savings > 0 or self.wallet > 0 or self.bank.bank_to_loan > 0:
            # create list of people at my location (includes self)
            my_cell = self.model.grid.get_cell_list_contents([self.pos])
            # check if other people are at my location
            if len(my_cell) > 1:
                # set customer to self for while loop condition
                customer = self
                while customer == self:
                    """select a random person from the people at my location
                    to trade with"""
                    customer = self.random.choice(my_cell)
                # 50% chance of trading with customer
                if self.random.randint(0, 1) == 0:
                    # 50% chance of trading $5
                    if self.random.randint(0, 1) == 0:
                        # give customer $5 from my wallet (may result in negative wallet)
                        customer.wallet += 5
                        self.wallet -= 5
                    # 50% chance of trading $2
                    else:
                        # give customer $2 from my wallet (may result in negative wallet)
                        customer.wallet += 2
                        self.wallet -= 2

    def balance_books(self):
        # check if wallet is negative from trading with customer
        if self.wallet < 0:
            # if negative money in wallet, check if my savings can cover the balance
            if self.savings >= (self.wallet * -1):
                """if my savings can cover the balance, withdraw enough
                money from my savings so that my wallet has a 0 balance"""
                self.withdraw_from_savings(self.wallet * -1)
            # if my savings cannot cover the negative balance of my wallet
            else:
                # check if i have any savings
                if self.savings > 0:
                    """if i have savings, withdraw all of it to reduce my
                    negative balance in my wallet"""
                    self.withdraw_from_savings(self.savings)
                # record how much money the bank can loan out right now
                temp_loan = self.bank.bank_to_loan
                """check if the bank can loan enough money to cover the
                   remaining negative balance in my wallet"""
                if temp_loan >= (self.wallet * -1):
                    """if the bank can loan me enough money to cover
                    the remaining negative balance in my wallet, take out a
                    loan for the remaining negative balance"""
                    self.take_out_loan(self.wallet * -1)
                else:
                    """if the bank cannot loan enough money to cover the negative
                    balance of my wallet, then take out a loan for the
                    total amount the bank can loan right now"""
                    self.take_out_loan(temp_loan)
        else:
            """if i have money in my wallet from trading with customer, deposit
            it to my savings in the bank"""
            self.deposit_to_savings(self.wallet)
        # check if i have any outstanding loans, and if i have savings
        if self.loans > 0 and self.savings > 0:
            # check if my savings can cover my outstanding loans
            if self.savings >= self.loans:
                # payoff my loans with my savings
                self.withdraw_from_savings(self.loans)
                self.repay_a_loan(self.loans)
            # if my savings won't cover my loans
            else:
                # pay off part of my loans with my savings
                self.withdraw_from_savings(self.savings)
                self.repay_a_loan(self.wallet)
        # calculate my wealth
        self.wealth = self.savings - self.loans

    # part of balance_books()
    def deposit_to_savings(self, amount):
        # take money from my wallet and put it in savings
        self.wallet -= amount
        self.savings += amount
        # increase bank deposits
        self.bank.deposits += amount

    # part of balance_books()
    def withdraw_from_savings(self, amount):
        # put money in my wallet from savings
        self.wallet += amount
        self.savings -= amount
        # decrease bank deposits
        self.bank.deposits -= amount

    # part of balance_books()
    def repay_a_loan(self, amount):
        # take money from my wallet to pay off all or part of a loan
        self.loans -= amount
        self.wallet -= amount
        # increase the amount the bank can loan right now
        self.bank.bank_to_loan += amount
        # decrease the bank's outstanding loans
        self.bank.bank_loans -= amount

    # part of balance_books()
    def take_out_loan(self, amount):
        """borrow from the bank to put money in my wallet, and increase my
        outstanding loans"""
        self.loans += amount
        self.wallet += amount
        # decresae the amount the bank can loan right now
        self.bank.bank_to_loan -= amount
        # increase the bank's outstanding loans
        self.bank.bank_loans += amount

    # step is called for each agent in model.BankReservesModel.schedule.step()
    def step(self):
        # move to a cell in my Moore neighborhood
        self.random_move()
        # trade
        self.do_business()
        # deposit money or take out a loan
        self.balance_books()
        # update the bank's reserves and the amount it can loan right now
        self.bank.bank_balance()



import copy

# from aiondao.utils import networking as netty
from dataclasses import dataclass as dataclassog
from enum import Enum, auto
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from aiondao import config
from aiondao._imports import *
from aiondao.agents.commons import TokenBatch
from aiondao.interfaces.models import IEconomy
from aiondao.libs.conviction import trigger_threshold
from aiondao.libs.states import SimState
from aiondao.schemas import Defector
from aiondao.utils._utilz import attrs
from mesa import Agent, Model

RandomNumCallable = Callable[..., float]
RandomNumCallable = Callable

# from aiondao.libs.graph import GraphORM
CHANGE_DELTA = {
    "failed": config.sentiment_bonus_proposal_becomes_failed,
    "success": config.sentiment_bonus_proposal_becomes_completed,
}


class StateAccessors:
    def __init__(self) -> None:
        self.__state = None

    # Properties to access network, functinf pool, and graphorm queries
    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, state: SimState):
        self.__state = state

    @property
    def network(self):
        return self.__state.network

    @property
    def capt(self):
        return self.__state.confs

    @property
    def cp(self):
        return self.__state.confs

    @property
    def commons(self):
        return self.__state.commons

    @property
    def gx(self):
        return GraphORM(net=self.__state.network)

    @property
    def funding_pool(self):
        return self.__state.funding_pool

    @property
    def funding_pool(self):
        return self.__state.token_supply

    @property
    def gamma_func(self):
        return self.__state.confs.gamma_func

    @property
    def participants(self):
        return self.gx.parts.all()

    @property
    def proposals(self):
        return self.gx.props.all()

    @property
    def exp_fn(self):
        return self.__state.confs.exponential_func

    @property
    def prob_fn(self):
        return self.__state.confs.exponential_func

    def register_state(self, state: SimState):
        self.state = state


class ProposalStatus(Enum):
    CANDIDATE = auto()
    ACTIVE = auto()
    COMPLETED = auto()
    FAILED = auto()


class ParticipantSupport(SyncModel):
    affinity: float
    tokens: float = 0.0
    conviction: float = 0.0
    is_author: bool = False


# candidate: proposal is being evaluated by the commons
# active: has been approved and is funded
# completed: the proposal was effective/successful
# failed: did not get to active status or failed after funding


@dataclass(config=autovalid)
class Proposal:
    # Fields that are required
    funds_requested: int
    trigger: float

    conviction = 0
    status = ProposalStatus.CANDIDATE
    age = 0

    def update_age(self):
        self.age += 1
        return self.age

    def update_threshold(
        self, funding_pool: float, token_supply: float, max_proposal_request: float
    ):
        if self.status == ProposalStatus.CANDIDATE:
            self.trigger = trigger_threshold(
                self.funds_requested, funding_pool, token_supply, max_proposal_request
            )
        else:
            self.trigger = np.nan
        return self.trigger

    def has_enough_conviction(
        self, funding_pool: float, token_supply: float, max_proposal_request
    ):
        """
        It's just a conviction < threshold check, but we recalculate the
        trigger_threshold so that the programmer doesn't have to remember to run
        update_threshold before running this method.
        """
        if self.status == ProposalStatus.CANDIDATE:
            threshold = trigger_threshold(
                self.funds_requested, funding_pool, token_supply, max_proposal_request
            )
            if self.conviction < threshold:
                return False
            return True
        else:
            raise (
                Exception(
                    "Proposal is not a Candidate Proposal and so asking it if it will pass is inappropriate"
                )
            )


class ParticipantDecision(StateAccessors):
    def __init__(self, participant: "Participant") -> None:
        super().__init__()
        self.participant = participant
        # self.__state: Optional[SimState] = None

    @property
    def model(self):
        return self.participant.model

    # @property
    # def net(self):
    #     """The net property."""
    #     if self.__state is None:
    #         raise AttributeError("State space was not found.")  # type: ignore
    #     return self.__state.network

    @property
    def commons(self):
        """The net property."""
        if self.__state is None:
            raise AttributeError("State space was not found.")  # type: ignore
        return self.__state.commons

    def is_state(self) -> bool:
        return True if self.__state else False

    def reset(self, state: SimState, params: Dict[str, Any] = {}):
        self.aggregate = 0.0
        self.__state = state

    """Processing Buy Decisions"""

    def decide_buy(self) -> Tuple[bool, float]:
        buy_amount = self.participant.buy()
        return bool(buy_amount), buy_amount

    def process_buy(self):
        is_buy, total_dai = self.decide_buy()
        commons2 = copy.copy(self.commons)
        if not is_buy:
            return
        # Calculate the amount minted that'll likely be minted
        estimated_tokens, estimated_token_price = commons2.deposit(total_dai)
        if not bool(total_dai):
            return

        minted_tokens, realized_price = self.commons.deposit(total_dai)
        if estimated_tokens != minted_tokens or estimated_token_price != realized_price:
            raise Exception(
                "ParticipantBuysTokens: {} tokens were minted at a price of {} (expected: {} with price {})".format(
                    minted_tokens,
                    realized_price,
                    estimated_tokens,
                    estimated_token_price,
                )
            )
        self.participant.increase_holdings(minted_tokens)

    """Processing Sell Decisions"""

    def decide_sell(self) -> Tuple[bool, float]:
        sell_amount = self.participant.sell()
        return bool(sell_amount), sell_amount

    def estimate_sale(self, tokens_sold: float):
        commons2 = copy.copy(self.commons)
        # Calculate the amount minted that'll likely be minted
        return commons2.burn(tokens_sold)

    def process_sell(self):
        is_buy, tokens_sold = self.decide_sell()
        if not is_buy:
            return

        # Estimate the remaining price and and dai returned
        estimated_dai, estimated_price = self.estimate_sale(tokens_sold)

        if not bool(tokens_sold):
            logger.info("ParticipantSellsTokens: did not sell tokens")
            return

        dai_returned, realized_price = self.commons.burn(tokens_sold)
        if dai_returned != estimated_dai or estimated_price != realized_price:
            raise Exception(
                "ParticipantSellsTokens: {} tokens were minted at a price of {} (expected: {} with price {})".format(
                    tokens_sold,
                    dai_returned,
                    realized_price,
                )
            )
        self.participant.spend(tokens_sold)

    """Dealing with Participant Exits"""

    def participant_decides_exit(self) -> bool:
        if not self.participant.wants_to_exit():
            return False
        return True

    # Function to get participants sentiment and holdings
    def get_participant_info(self) -> Dict[str, Any]:
        return Defector(
            sentiment=self.participant.sentiment,
            holdings=self.participant.holdings.total,
        )

    def burn_exiters_tokens(self, exiter_holdings: float):
        dai_returned, realized_price = self.commons.burn(exiter_holdings)

    def process_exit_logic(self):
        if not self.participant_decides_exit():
            # logger.info(
            #     f"Participant {self.participant.unique_id} does not want to exit"
            # )
            return
        logger.error(f"Participant {self.participant.unique_id} wants to exit")
        defect = self.get_participant_info()
        self.network.remove_node(self.participant.unique_id)
        self.participant.has_existed = True
        self.burn_exiters_tokens(defect.holdings)

    def get_proposals(self, status_types: List[ProposalStatus] = []) -> List[Proposal]:
        from aiondao.utils.networking import get_proposals_by_participant_and_status

        if not status_types:
            # Get all of the status types
            status_types = [
                ProposalStatus.ACTIVE,
                ProposalStatus.CANDIDATE,
                ProposalStatus.COMPLETED,
                ProposalStatus.FAILED,
            ]
        return get_proposals_by_participant_and_status(
            self.net, self.participant.unique_id, status_filter=status_types
        )

    def update_sentiment_conviction(self):
        logger.info(f"Updating active sentiment for {self.participant.unique_id}")
        convictions = self.state.conviction_proposals

        for prop_id in convictions:
            edge = self.network.edges[self.participant.unique_id, prop_id]
            if not "support" in edge:
                continue
            edge_support = edge["support"]
            if not edge_support.is_author:
                continue
            part = self.participant
            s1 = part.sentiment
            sentiment = np.clip(
                s1 + config.sentiment_bonus_proposal_becomes_active,
                a_min=0.0,
                a_max=1.0,
            )
            self.participant.sentiment = sentiment

    def update_sentiment_success_fail(self):
        logger.success(
            f"Updating success/fail sentiment for {self.participant.unique_id}"
        )
        for status, proposal_idxs in self.state.policy_output.items():
            if not proposal_idxs:
                continue
            for proposal_idx in proposal_idxs:
                edge = self.network.edges[self.participant.unique_id, proposal_idx]
                if not "support" in edge:
                    continue

                supporter = edge["support"]
                if not (supporter.is_author or supporter.tokens > 0):
                    continue

                logger.warning(supporter)
                s1 = self.participant.sentiment
                sentiment = np.clip(
                    s1 + (supporter.affinity * CHANGE_DELTA[status]),
                    a_min=0.0,
                    a_max=1.0,
                )
                self.participant.sentiment = sentiment

    def process_sentiment(self):
        # Go through proposal logic and update sentiment
        self.update_sentiment_conviction()
        self.update_sentiment_success_fail()
        # for idx in policy_output_passthru["proposal_idxs_with_enough_conviction"]:
        # self.participant.sentiment = self.participant.sentiment_update()


# @dataclass


class Participant(BorgIdModel, Agent):
    holdings: TokenBatch = None
    prob_fn: Callable = None
    random_number_func: Callable = None
    model: Optional[Union["Model", IEconomy]] = None
    unique_id: Optional[int] = None
    sentiment: float = 0
    has_existed: bool = False
    __step_logic: ParticipantDecision = None

    def __post_init__(self):
        self.prob_fn = self.prob_fn
        self._random_number_func = self.random_number_func

        self.sentiment = self._random_number_func()

    @property
    def logic(self) -> ParticipantDecision:
        if self.__step_logic is None:
            self.__step_logic = ParticipantDecision(self)
        return self.__step_logic

    @property
    def flags(self):
        return self.model.flags

    def buy(self) -> float:
        """
        If the Participant decides to buy more tokens, returns the number of
        tokens. Otherwise, return 0.

        This method does not modify itself, it simply returns the answer so that
        cadCAD's state update functions will make the changes and maintain its
        functional-ness.
        """
        engagement_rate = config.engagement_rate_multiplier_buy * self.sentiment
        force = self.sentiment - config.sentiment_sensitivity
        if self.prob_fn(engagement_rate) and force > 0:
            delta_holdings = (
                self.random_number_func() * force * config.delta_holdings_scale
            )
            return delta_holdings
        return 0

    def sell(self) -> float:
        """
        Decides to sell some tokens, and if so how many. If the Participant
        decides to sell some tokens, returns the number of tokens. Otherwise,
        return 0.

        This method does not modify itself, it simply returns the answer so that
        cadCAD's state update functions will make the changes and maintain its
        functional-ness.
        """
        engagement_rate = config.engagement_rate_multiplier_sell * self.sentiment
        force = self.sentiment - config.sentiment_sensitivity
        if self.prob_fn(engagement_rate) and force < 0:
            spendable = self.holdings.spendable()
            # It is expected that the function returns a positive value for the
            # amount sold.
            force = -1 * force
            delta_holdings = self.random_number_func() * force * spendable
            return delta_holdings
        return 0

    def increase_holdings(self, x: float):
        """
        increase_holdings() is the opposite of spend() and adds to the
        nonvesting part of the TokenBatch.
        """
        self.holdings.nonvesting += x
        return (
            self.holdings.vesting,
            self.holdings.vesting_spent,
            self.holdings.nonvesting,
        )

    def spend(self, x: float) -> Tuple[float, float, float]:
        """
        Participant.spend() is simply a front to TokenBatch.spend().
        """
        return self.holdings.spend(x)

    def create_proposal(
        self, total_funds_requested, median_affinity, funding_pool
    ) -> bool:
        """
        Here the Participant will decide whether or not to create a new
        Proposal.

        This equation, originally from randomly_gen_new_proposal(), is a
        systems-type simulation. An individual Participant would likely think in
        a different way, and thus this equation should change. Nevertheless for
        simplicity's sake, we use this same equation for now.

        Explanation: If the median affinity is high, the Proposal Rate should be
        high.

        If total funds_requested in candidate proposals is much lower than the
        funding pool (i.e. the Commons has lots of spare money), then people are
        just going to pour in more Proposals.
        """
        percent_of_funding_pool_being_requested = total_funds_requested / funding_pool
        proposal_rate = median_affinity / (1 + percent_of_funding_pool_being_requested)
        new_proposal = self.prob_fn(proposal_rate)
        return new_proposal

    def vote_on_candidate_proposals(self, candidate_proposals: dict) -> dict:
        """
        Here the Participant decides which Candidate Proposals he will stake
        tokens on. This method does not decide how many tokens he will stake
        on them, because another function should decide how the tokens should be
        balanced across the newly supported proposals and the ones the
        Participant already supported.

        Copied from
        participants_buy_more_if_they_feel_good_and_vote_for_proposals()

        candidate dict format:
        {
            proposal_idx: affinity,
            ...
        }

        NOTE: the original cadCAD policy returned {'delta_holdings':
        delta_holdings, 'proposals_supported': proposals_supported}
        proposals_supported seems to include proposals ALREADY supported by the
        participant, but I don't think it is needed.
        """
        new_voted_proposals = {}
        engagement_rate = 1.0
        if self.prob_fn(engagement_rate):
            # Put your tokens on your favourite Proposals, where favourite is
            # calculated as 0.75 * (the affinity for the Proposal you like the
            # most) e.g. if there are 2 Proposals that you have affinity 0.8,
            # 0.9, then 0.75*0.9 = 0.675, so you will end up voting for both of
            # these Proposals
            #
            # A Zargham work of art.
            for candidate in candidate_proposals:
                affinity = candidate_proposals[candidate]
                # Hardcoded 0.75 instead of a configurable sentiment_sensitivity
                # because modifying sentiment_sensitivity without changing the
                # hardcoded cutoff value of 0.5 may cause unintended behaviour.
                # Also, 0.75 is a reasonable number in this case.
                cutoff = config.candidate_proposals_cutoff * np.max(
                    list(candidate_proposals.values())
                )
                if cutoff < 0.5:
                    cutoff = 0.5

                if affinity > cutoff:
                    new_voted_proposals[candidate] = affinity

        return new_voted_proposals

    def stake_across_all_supported_proposals(
        self, supported_proposals: List[Tuple[float, int]]
    ) -> dict:
        """
        Rebalances the Participant's tokens across the (possibly updated) list of Proposals
        supported by this Participant.

        These tokens can come from a Participant's vesting and nonvesting TokenBatches.

        supported_proposals format:
        [(affinity, proposal_idx)...]
        """
        tokens_per_supported_proposal = {}
        supported_proposals = sorted(supported_proposals, key=lambda tup: tup[0])

        affinity_total = sum([a for a, idx in supported_proposals])
        for affinity, proposal_idx in supported_proposals:
            tokens_per_supported_proposal[proposal_idx] = self.holdings.total * (
                affinity / affinity_total
            )

        return tokens_per_supported_proposal

    def wants_to_exit(self):
        """
        Returns True if the Participant wants to exit (if sentiment < 0.5,
        random chance of exiting) and if the Participant has no vesting
        token, otherwise False.
        """
        sensitivity_exit = config.sentiment_sensitivity_exit
        vesting = self.holdings.vesting

        if self.sentiment < sensitivity_exit and vesting == 0:
            engagement_rate = config.engagement_rate_multiplier_exit * self.sentiment
            return self.prob_fn(1 - engagement_rate)
        return False

    def update_token_batch_age(self):
        """
        Participant.update_token_batch_age() is simply a front to
        TokenBatch.update_age().
        """
        return self.holdings.update_age()

    def step(self):
        # self.logic.reset(self.state)
        self.logic.register_state(self.model.state)
        # self.logic.reset(self.state)
        # ! Update Token Holdings if you haven't already
        self.logic.process_buy()
        self.logic.process_sell()
        self.logic.process_exit_logic()
        self.logic.process_sentiment()
        # is_buy, amount = self.logic.decide_buy()
        # debug((is_buy, amount))
        return super().step()



from aiondao._imports import *
from loguru import logger

from typing import List, Optional, Tuple

from mesa import Agent, Model
from aiondao.libs.abcurve import AugmentedBondingCurve

from aiondao import config
from pydantic import Extra, root_validator

from aiondao.interfaces.models import IEconomy

ilog = logger.info


def vesting_curve(day: int, cliff_days: int, halflife_days: float) -> float:
    """
    The vesting curve includes the flat cliff, and the halflife curve where tokens are gradually unlocked.
    It looks like _/--
    """
    return 1 - config.vesting_curve_halflife ** ((day - cliff_days) / halflife_days)


def convert_80p_to_cliff_and_halflife(
    days: int, v_ratio: int = 2
) -> Tuple[float, float]:
    """
    For user's convenience, we ask him after how many days he would like 80% of his tokens to be unlocked.
    This needs to be converted into a half life (unit days).
    2.321928094887362 is log(base0.5) 0.2, or log 0.2 / log 0.5.
    v_ratio is cliff / halflife, and its default is determined by Commons Stack
    """
    halflife_days = days / (config.log_base05_of_02 + v_ratio)
    cliff_days = v_ratio * halflife_days
    return cliff_days, halflife_days


def hatch_raise_split_pools(total_hatch_raise, hatch_tribute) -> Tuple[float, float]:
    """Splits the hatch raise between the funding / collateral pool based on the fraction."""
    funding_pool = hatch_tribute * total_hatch_raise
    collateral_pool = total_hatch_raise * (1 - hatch_tribute)
    return funding_pool, collateral_pool


@dataclass(config=autovalid, order=True)
class VestingOptions:
    cliff_days: int
    halflife_days: int


@dataclass(config=autovalid)
class TokenBatch:
    vesting: float
    nonvesting: float
    vesting_spent: float = 0.0
    age_days: int = 0

    vesting_options: Optional[VestingOptions] = None

    def __post_init__(self):
        self.vesting_options = cast(VestingOptions, self.vesting_options)
        self.cliff_days = (
            0 if not self.vesting_options else self.vesting_options.cliff_days  # type: ignore
        )
        self.halflife_days = (
            0 if not self.vesting_options else self.vesting_options.halflife_days  # type: ignore
        )

    @property
    def total(self) -> float:
        return (self.vesting - self.vesting_spent) + self.nonvesting

    def __bool__(self):
        if self.total > 0:
            return True
        return False

    def __add__(self, other):
        total_vesting = self.vesting + other.vesting
        total_nonvesting = self.nonvesting + other.nonvesting
        return total_vesting, total_nonvesting

    def __sub__(self, other):
        total_vesting = self.vesting - other.vesting
        total_nonvesting = self.nonvesting - other.nonvesting
        return total_vesting, total_nonvesting

    def update_age(self, iterations: int = 1):
        """
        Adds the number of iterations to TokenBatch.age_days
        """
        self.age_days += iterations
        return self.age_days

    def unlocked_fraction(self) -> float:
        """
        returns what fraction of the TokenBatch is unlocked to date
        """
        if self.cliff_days and self.halflife_days:
            u = vesting_curve(self.age_days, self.cliff_days, self.halflife_days)
            return u if u > 0 else 0
        else:
            return 1.0

    def spend(self, x: float):
        """
        checks if you can spend so many tokens, then decreases this TokenBatch
        instance's value accordingly
        """
        if x > self.spendable():
            raise Exception(
                "Not so many tokens are available for you to spend yet ({})".format(
                    self.age_days
                )
            )

        y = x - self.nonvesting
        if y > 0:
            self.vesting_spent += y
            self.nonvesting = 0.0
        else:
            self.nonvesting = abs(y)

        return self.vesting, self.vesting_spent, self.nonvesting

    def spendable(self) -> float:
        """
        spendable() = (self.unlocked_fraction * self.vesting - self.vesting_spent) + self.nonvesting
        Needed in case some Tokens were burnt before.
        """
        return (
            (self.unlocked_fraction() * self.vesting) - self.vesting_spent
        ) + self.nonvesting


def create_token_batches(
    hatcher_contributions: List[float],
    desired_token_price: float,
    cliff_days: float,
    halflife_days: float,
) -> Tuple[List[TokenBatch], float]:
    """
    hatcher_contributions: a list of hatcher contributions in DAI/ETH/whatever
    desired_token_price: used to determine the initial token supply
    vesting_80p_unlocked: vesting parameter - the number of days after which 80% of tokens will be unlocked, including the cliff period
    """
    total_hatch_raise = sum(hatcher_contributions)
    initialtoken_supply = total_hatch_raise / desired_token_price

    # In the hatch, everyone buys in at the same time, with the same price. So just split the token supply amongst the hatchers proportionally to their contributions
    tokens_per_hatcher = [
        (x / total_hatch_raise) * initialtoken_supply for x in hatcher_contributions
    ]

    token_batches = [
        TokenBatch(x, 0, vesting_options=VestingOptions(cliff_days, halflife_days))
        for x in tokens_per_hatcher
    ]
    return token_batches, initialtoken_supply


class CommonAgent(Agent, IEconomy):
    def __init__(
        self,
        unique_id: int,
        model: Model,
        total_hatch_raise: float,
        token_supply: float,
        *args,
        hatch_tribute: float = 0.2,
        exit_tribute: float = 0,
        kappa: int = 2,
        hatch_tokens: Optional[float] = None,
        _funding_pool: Optional[float] = None,
        _collateral_pool: Optional[float] = None,
        bonding_curve: Optional[AugmentedBondingCurve] = None,
        **kwargs,
    ) -> None:
        super().__init__(unique_id, model, *args, **kwargs)

        self.total_hatch_raise = total_hatch_raise
        self.token_supply = token_supply
        self.hatch_tribute = hatch_tribute
        self.exit_tribute = exit_tribute
        self.kappa = kappa
        self.hatch_tokens = hatch_tokens

        self._collateral_pool = (1 - self.hatch_tribute) * self.total_hatch_raise
        self._funding_pool = (
            self.hatch_tribute * self.total_hatch_raise
        )  # 0.35 * total_hatch_raise = 35%

        # hatch_tokens keeps track of the number of tokens that were created when hatching, so we can calculate the unlocking of those
        self.hatch_tokens = self.token_supply
        self.bonding_curve = AugmentedBondingCurve(
            reserve=self._collateral_pool,
            supply=self.token_supply,
            kappa=self.kappa,
        )

        # Options
        self.exit_tribute = self.exit_tribute

    def __repr__(self) -> str:
        return f"Commons(collateral={self.collateral_pool}, funding={self.funding_pool}, tokens={self.token_supply})"

    @property
    def funding_pool(self) -> float:
        """The funding_pool property."""
        return self._funding_pool

    @funding_pool.setter
    def funding_pool(self, value: Union[float, int]):
        self._funding_pool = value

    @property
    def collateral_pool(self) -> float:
        """The collateral_pool property."""
        return self._collateral_pool

    @collateral_pool.setter
    def collateral_pool(self, value: float):
        self._collateral_pool = value

    @property
    def _token_supply(self) -> float:
        """The _token_supply property."""
        return self.token_supply

    @_token_supply.setter
    def _token_supply(self, value: float):
        self.token_supply = value

    def deposit(self, dai):
        """
        Deposit DAI after the hatch phase. This means all the incoming deposit goes to the collateral pool.
        """
        tokens, realized_price = self.bonding_curve.deposit(
            dai, self._collateral_pool, self.token_supply
        )
        self.token_supply += tokens
        self._collateral_pool += dai
        return tokens, realized_price

    def burn(self, tokens):
        """
        Burn tokens, with/without an exit tribute.
        """
        dai, realized_price = self.bonding_curve.burn(
            tokens, self._collateral_pool, self.token_supply
        )
        self.token_supply -= tokens
        self._collateral_pool -= dai
        money_returned = dai

        if self.exit_tribute:
            self._funding_pool += self.exit_tribute * dai
            money_returned = (1 - self.exit_tribute) * dai
        return money_returned, realized_price

    def dai_to_tokens(self, dai):
        """
        Given the size of the common's collateral pool, return how many tokens would x DAI buy you.
        """
        price = self.bonding_curve.get_token_price(self._collateral_pool)
        return dai / price

    def token_price(self):
        """
        Query the bonding curve for the current token price, given the size of the commons's collateral pool.
        """
        return self.bonding_curve.get_token_price(self._collateral_pool)

    def spend(self, amount):
        """
        Decreases the Common's funding_pool by amount.
        Raises an exception if this would make the funding pool negative.
        """
        if self._funding_pool - amount < 0:
            raise Exception(
                "{} funds requested but funding pool only has {}".format(
                    amount, self._funding_pool
                )
            )
        self._funding_pool -= amount
        return

    def start_step(self):
        pass
        # return super().start_step()

    def end_step(self):
        pass

    def step(self) -> None:

        return super().step()


class LooseWrapper(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow


class Common(LooseWrapper):
    # All of these are to be calculated at a later time.
    total_hatch_raise: float
    token_supply: float
    hatch_tribute: float = 0.2
    exit_tribute: float = 0.0
    kappa: float = 2

    hatch_tokens: Optional[float] = None
    _funding_pool: Optional[float] = None
    _collateral_pool: Optional[float] = None
    # bonding_curve: AugmentedBondingCurve

    @root_validator
    def validate_commons(cls, values):
        logger.warning("Validating commons")
        values["_collateral_pool"] = (1 - values["hatch_tribute"]) * values[
            "total_hatch_raise"
        ]
        values["_funding_pool"] = (
            values["hatch_tribute"] * values["total_hatch_raise"]
        )  # 0.35 * total_hatch_raise = 35%

        kappa = values.get("kappa", 2)
        collateral = values["_collateral_pool"]
        supply = values["token_supply"]

        logger.info((kappa, collateral, supply))
        # hatch_tokens keeps track of the number of tokens that were created when hatching, so we can calculate the unlocking of those
        # values["hatch_tokens"] = values["token_supply"]
        values["bonding_curve"] = AugmentedBondingCurve(
            reserve=collateral,
            supply=supply,
            kappa=kappa,
        )

        # Options
        values["exit_tribute"] = values.get("exit_tribute", 0.2)

        return values

    @property
    def _token_supply(self) -> float:
        """The _token_supply property."""
        return self.token_supply

    @_token_supply.setter
    def _token_supply(self, value: float):
        self.token_supply = value

    @property
    def collateral_pool(self) -> float:
        """The collateral_pool property."""
        if self._collateral_pool is not None:
            return self._collateral_pool
        self._collateral_pool = (1 - self.hatch_tribute) * self.total_hatch_raise
        return self._collateral_pool

    @collateral_pool.setter
    def collateral_pool(self, value: float):
        self._collateral_pool = value

    def deposit(self, dai):
        """
        Deposit DAI after the hatch phase. This means all the incoming deposit goes to the collateral pool.
        """
        tokens, realized_price = self.bonding_curve.deposit(
            dai, self.collateral_pool, self.token_supply
        )
        self.token_supply += tokens
        self._collateral_pool += dai
        return tokens, realized_price

    def burn(self, tokens):
        """
        Burn tokens, with/without an exit tribute.
        """

        dai, realized_price = self.bonding_curve.burn(
            tokens, self.collateral_pool, self.token_supply
        )
        self.token_supply -= tokens
        self._collateral_pool -= dai
        money_returned = dai

        if self.exit_tribute:
            self._funding_pool += self.exit_tribute * dai
            money_returned = (1 - self.exit_tribute) * dai
        return money_returned, realized_price

    def dai_to_tokens(self, dai):
        """
        Given the size of the common's collateral pool, return how many tokens would x DAI buy you.
        """
        price = self.bonding_curve.get_token_price(self.collateral_pool)
        return dai / price

    def token_price(self):
        """
        Query the bonding curve for the current token price, given the size of the commons's collateral pool.
        """
        return self.bonding_curve.get_token_price(self.collateral_pool)

    def spend(self, amount):
        """
        Decreases the Common's funding_pool by amount.
        Raises an exception if this would make the funding pool negative.
        """
        if self._funding_pool - amount < 0:
            raise Exception(
                "{} funds requested but funding pool only has {}".format(
                    amount, self._funding_pool
                )
            )
        self._funding_pool -= amount
        return

    def step(self) -> None:
        return super().step()

    # def update_bond(self):
    #     self.bonding_curve.reserve = self._collateral_pool

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow


#
Commons = Common


def main():
    pass


if __name__ == "__main__":
    main()



import copy
from typing import cast
from aiondao.libs.states import SimState

import numpy as np
from aiondao import config
from aiondao.agents.commons import TokenBatch
from aiondao.agents.members import Participant
from aiondao.utils.networking import (
    add_participant,
    find_in_edges_of_type_for_proposal,
    get_participants,
)
from auto_all import start_all
from aiondao.schemas import (
    InvestmentInput,
    PolicyAggregate,
    ParticipantPurchase,
)

from aiondao.libs.states import SimState
from aiondao._imports import *


# from aiondao.interfaces

start_all(globals())


class CreateNewParticipant:
    def __init__(self):
        self.agg = PolicyAggregate()

    def randomize(self, state: SimState):
        commons = state.commons
        fns = state.to_fns()
        sentiment = state.sentiment
        timestep = state.timestep

        speculation_days = state.confs.speculation_days
        multiplier_new_participants = state.confs.multiplier_new_participants

        dict_ans = {}
        if timestep < speculation_days:
            # If in speculation period, the arrival rate is higher
            arrival_rate = 0.5 + 0.5 * sentiment
            multiplier = multiplier_new_participants
            max_new_participants = config.max_new_participants * multiplier
        else:
            arrival_rate = (1 + sentiment) / config.arrival_rate_denominator
            max_new_participants = config.max_new_participants

        for _ in range(int(max_new_participants)):
            if fns.prob_fn(arrival_rate):
                # Here we randomly generate each participant's post-Hatch
                # investment, in DAI/USD.
                #
                # expon.rvs() arguments:
                #
                # loc is the minimum number, so if loc=100, there will be no
                # investments < 100
                #
                # scale is the standard deviation, so if scale=2, investments will
                # be around 0-12 DAI or even 15, if scale=100, the investments will be
                # around 0-600 DAI.
                participant_investment = fns.exp_fn(
                    loc=config.investment_new_participant_min,
                    scale=config.investment_new_participant_stdev,
                )
                tokens = commons.dai_to_tokens(participant_investment)

                self.agg.invests.append(
                    InvestmentInput(investment=participant_investment, tokens=tokens)
                )
        return dict_ans

    def su_add_to_network(self, state: SimState):

        network = state.network

        probability_func = state.confs.probability_func
        exponential_func = state.confs.exponential_func
        random_number_func = state.confs.random_number_func
        # speculation_days = state.confs.random_func
        # network = state.network
        # probability_func = params["probability_func"]
        # exponential_func = params["exponential_func"]
        random_number_func = state.confs.random_number_func
        params = {}
        for idx, invest in enumerate(self.agg.invests):
            if invest.is_valid():
                network, idx = add_participant(
                    network,
                    Participant(
                        holdings=TokenBatch(vesting=0, nonvesting=invest.tokens),
                        prob_fn=probability_func,  # type: ignore
                        random_number_func=random_number_func,  # type: ignore
                    ),
                    exponential_func,
                    random_number_func,
                )

                if params.get("debug", True):
                    print(
                        "GenerateNewParticipant: A new Participant {} invested {}DAI for {} tokens".format(
                            idx, invest.investment, invest.tokens
                        )
                    )
        state.network = network
        return "network", network

    def su_add_investment_to_commons(self, state: SimState):
        commons = state.commons
        for _, invest in enumerate(self.agg.invests):

            if invest.is_valid():
                commons.deposit(invest.investment)
        state.commons = commons
        return "commons", commons

    def su_update_participants_token_batch_age(self, state: SimState, **kwargs):
        network = state.network
        participants = get_participants(network)
        for pidx, participant in participants:
            participant = cast(Participant, participant)

            participant.update_token_batch_age()
        state.network = network
        return "network", network


class ParticipantBuysTokens:
    """
    When implementing buying tokens, Participants may either buy in bulk with no
    slippage, or endure slippage while buying. Neither is realistic but I'd say
    no slippage is closer to what Participants will experience in real life.

    When implementing selling tokens, slippage is more likely to change the
    decision. In any case, let us implement no slippage first since it is easier
    to reason with.

    Also, buying in bulk is easier to implement in cadCAD because if the state
    update function changes the Commons object each time a Participant buys
    in/sells out, then it would have to update the network and commons object in
    the same function, which is not allowed in cadCAD.
    """

    def p_decide_to_buy_tokens_bulk(self, state: SimState):
        network = state.network
        commons = state.commons
        participants = get_participants(network)
        ans = {}
        params = {}
        total_dai = 0

        for i, participant in participants:
            participant = cast(Participant, participant)
            # If a participant decides to buy, it will be specified in units of DAI.
            # If a participant decides to sell, it will be specified in units of tokens.
            x = participant.buy()
            if x > 0:
                total_dai += x
                ans[i] = x

        # Now that we have the sum of DAI, ask the Commons object how many
        # tokens this would be minted as a result. This will be inaccurate due
        # to slippage, and we need the result of this policy to be final to
        # avoid chaining 2 state update functions, so we run the deposit on a
        # throwaway copy of Commons
        if total_dai == 0:
            if params.get("debug", True):
                print(
                    "ParticipantBuysTokens: No Participants bought tokens in timestep {}".format(
                        state.timestep
                    )
                )
            return ParticipantPurchase(
                **{
                    "participant_decisions": ans,
                    "total_dai": 0,
                    "tokens": 0,
                    "token_price": 0,
                    "final_token_distribution": {},
                }
            )

        else:
            commons2 = copy.copy(commons)
            tokens, token_price = commons2.deposit(total_dai)

            final_token_distribution = {}
            for i in ans:
                final_token_distribution[i] = ans[i] / total_dai

            if params.get("debug", True):
                print(
                    "ParticipantBuysTokens: These Participants have decided to buy tokens with this amount of DAI: {}".format(
                        ans
                    )
                )
                print(
                    "ParticipantBuysTokens: A total of {} DAI will be deposited. {} tokens should be minted as a result at a price of {} DAI/token".format(
                        total_dai, tokens, token_price
                    )
                )

            return ParticipantPurchase(
                **{
                    "participant_decisions": ans,
                    "total_dai": total_dai,
                    "tokens": tokens,
                    "token_price": token_price,
                    "final_token_distribution": final_token_distribution,
                }
            )

    def su_buy_participants_tokens(
        self, state: SimState, agg: PolicyAggregate, **kwargs
    ):
        commons = state.commons
        purchase = agg.purchase
        if purchase is None:
            return
        if purchase.total_dai > 0:
            tokens, realized_price = commons.deposit(purchase.total_dai)
            if purchase.tokens != tokens or purchase.token_price != realized_price:
                raise Exception(
                    "ParticipantBuysTokens: {} tokens were minted at a price of {} (expected: {} with price {})".format(
                        tokens, realized_price, purchase.tokens, purchase.token_price
                    )
                )

        return "commons", commons

    def su_update_participants_tokens(
        self, state: SimState, agg: PolicyAggregate, **kwargs
    ):
        network = state.network
        purchase = agg.purchase
        if purchase is None:
            return
        decisions = purchase.participant_decisions
        final_token_distribution = purchase.final_token_distribution
        tokens = purchase.tokens

        for participant_idx, decision in decisions.items():
            network.nodes[participant_idx]["item"].increase_holdings(
                final_token_distribution[participant_idx] * tokens
            )

        return "network", network


class ParticipantSellsTokens:
    def __init__(self) -> None:
        self.agg = PolicyAggregate()

    def p_decide_to_sell_tokens_bulk(self, state: SimState, **kwargs):
        network = state.network
        commons = state.commons
        participants = get_participants(network)
        ans = {}
        total_tokens = 0
        for i, participant in participants:
            # If a participant decides to buy, it will be specified in units of DAI.
            # If a participant decides to sell, it will be specified in units of tokens.
            x = participant.sell()
            if x > 0:
                total_tokens += x
                ans[i] = x

        # Now that we have the sum of tokens, ask the Commons object how many
        # DAI would be redeemed as a result. This will be inaccurate due
        # to slippage, and we need the result of this policy to be final to
        # avoid chaining 2 state update functions, so we run the operation on a
        # throwaway copy of Commons
        if total_tokens == 0:
            if params.get("debug"):
                print(
                    "ParticipantSellsTokens: No Participants sold tokens in timestep {}".format(
                        step
                    )
                )
            return PolicyAggregate(
                **{
                    "participant_decisions": ans,
                    "total_tokens": 0,
                    "dai_returned": 0,
                    "realized_price": 0,
                }
            )
        else:
            commons2 = copy.copy(commons)
            dai_returned, realized_price = commons2.burn(total_tokens)

            final_dai_distribution = {}
            for i in ans:
                final_dai_distribution[i] = ans[i] / total_tokens

            if params.get("debug"):
                print(
                    "ParticipantSellsTokens: These Participants have decided to sell this many  tokens: {}".format(
                        ans
                    )
                )
                print(
                    "ParticipantSellsTokens: A total of {} tokens will be burned. {} DAI should be returned as a result, at a price of {} DAI/token".format(
                        total_tokens, dai_returned, realized_price
                    )
                )

            return PolicyAggregate(
                **{
                    "participant_decisions": ans,
                    "total_tokens": total_tokens,
                    "dai_returned": dai_returned,
                    "realized_price": realized_price,
                }
            )

    def su_burn_participants_tokens(
        self, state: SimState, agg: PolicyAggregate, **kwargs
    ):
        commons = state.commons
        self.agg.purchase
        if agg.total_tokens > 0:
            dai_returned, realized_price = commons.burn(agg.total_tokens)
            if agg.dai_returned != dai_returned or agg.realized_price != realized_price:
                raise Exception(
                    "ParticipantSellsTokens: {} DAI was returned at a price of {} (expected: {} with price {})".format(
                        dai_returned,
                        realized_price,
                        agg.dai_returned,
                        agg.realized_price,
                    )
                )

        return "commons", commons

    def su_update_participants_tokens(
        self, state: SimState, aggregate: PolicyAggregate, **kwargs
    ):
        network = state.network
        purchase = aggregate.purchase
        if purchase is None:
            return
        decisions = purchase.participant_decisions
        dai_returned = purchase.dai_returned

        for participant_idx, decision in decisions.items():
            network.nodes[participant_idx]["item"].spend(decision)

        return "network", network


class ParticipantExits:
    def p_participant_decides_if_he_wants_to_exit(
        self, state: SimState, aggregate: PolicyAggregate, **kwargs
    ):
        # A lot of these decisions can ge moved to individual participants
        params = {}
        network = state.network
        participants = get_participants(network)
        defectors = {}
        for i, participant in participants:
            participant = cast(Participant, participant)
            if participant.wants_to_exit():
                defectors[i] = {
                    "sentiment": participant.sentiment,
                    "holdings": participant.holdings.total,
                }

        if params.get("debug"):
            print(
                "ParticipantExits: Participants {} (2nd number is their sentiment) want to exit".format(
                    defectors
                )
            )
        return {"defectors": defectors}

    def su_remove_participants_from_network(self, state: SimState, **kwargs):
        # network = state.network
        network = state.network
        defectors = _input["defectors"]

        for i, _ in defectors.items():
            network.remove_node(i)

        return "network", network

    def su_burn_exiters_tokens(self, state: SimState, **kwargs):
        commons = state.commons
        network = state.network
        defectors = _input["defectors"]

        burnt_token_results = {}
        for i, v in defectors.items():
            dai_returned, realized_price = commons.burn(v["holdings"])
            burnt_token_results[i] = {
                "dai_returned": dai_returned,
                "realized_price": realized_price,
            }

        return "commons", commons

    def su_update_sentiment_when_proposal_becomes_active(
        self, state: SimState, **kwargs
    ):
        network = state.network
        policy_output_passthru = state.policy_output
        if not policy_output_passthru:
            return

        report = {}
        for idx in policy_output_passthru["proposal_idxs_with_enough_conviction"]:
            for participant_idx, proposal_idx, _ in find_in_edges_of_type_for_proposal(
                network, idx, "support"
            ):
                edge = network.edges[participant_idx, proposal_idx]
                if edge["support"].is_author:
                    sentiment_old = network.nodes[participant_idx]["item"].sentiment
                    sentiment_new = (
                        sentiment_old + config.sentiment_bonus_proposal_becomes_active
                    )
                    sentiment_new = 1 if sentiment_new > 1 else sentiment_new
                    network.nodes[participant_idx]["item"].sentiment = sentiment_new

                    report[participant_idx] = {
                        "proposal_idx": proposal_idx,
                        "sentiment_old": sentiment_old,
                        "sentiment_new": sentiment_new,
                    }

        if params.get("debug"):
            for i in report:
                print(
                    "ParticipantExits: Participant {} changed his sentiment from {} to {} because Proposal {} became active".format(
                        i,
                        report[i]["sentiment_old"],
                        report[i]["sentiment_new"],
                        report[i]["proposal_idx"],
                    )
                )
        return "network", network

    def su_update_sentiment_when_proposal_becomes_failed_or_completed(
        self, state: SimState, **kwargs
    ):
        network = state.network
        policy_output_passthru = s["policy_output"]

        proposal_status_delta = {
            "failed": config.sentiment_bonus_proposal_becomes_failed,
            "succeeded": config.sentiment_bonus_proposal_becomes_completed,
        }
        report = {}
        for status, delta in proposal_status_delta.items():
            for idx in policy_output_passthru[status]:
                for (
                    participant_idx,
                    proposal_idx,
                    _,
                ) in find_in_edges_of_type_for_proposal(network, idx, "support"):
                    edge = network.edges[participant_idx, proposal_idx]
                    # Update the participant sentiment if he/she is the proposal creator
                    # or if participant has staked on the proposal (tokens > 0)
                    if edge["support"].is_author or edge["support"].tokens > 0:
                        sentiment_old = network.nodes[participant_idx]["item"].sentiment
                        sentiment_new = sentiment_old + (
                            edge["support"].affinity * delta
                        )
                        sentiment_new = np.clip(sentiment_new, a_min=0.0, a_max=1.0)
                        network.nodes[participant_idx]["item"].sentiment = sentiment_new

                        report[participant_idx] = {
                            "proposal_idx": proposal_idx,
                            "sentiment_old": sentiment_old,
                            "sentiment_new": sentiment_new,
                            "status": status,
                        }

        if params.get("debug"):
            for i in report:
                print(
                    "ParticipantExits: Participant {} changed his sentiment from {} to {} because Proposal {} became {}".format(
                        i,
                        report[i]["sentiment_old"],
                        report[i]["sentiment_new"],
                        report[i]["proposal_idx"],
                        report[i]["status"],
                    )
                )
        return "network", network


class ParticipantSentiment:
    def su_update_sentiment_decay(self, state: SimState, **kwargs):
        network = state.network

        participants = get_participants(network)
        for participant_idx, participant in participants:
            sentiment_old = network.nodes[participant_idx]["item"].sentiment
            sentiment_new = sentiment_old - config.sentiment_decay
            sentiment_new = 0 if sentiment_new < 0 else sentiment_new
            network.nodes[participant_idx]["item"].sentiment = sentiment_new

        return "network", network



from aiondao._imports import *

from typing import Optional, Tuple
import numpy as np
import copy

from pydantic import Field
from aiondao.agents.commons import (
    create_token_batches,
    Commons,
    convert_80p_to_cliff_and_halflife,
)

from aiondao.agents.members import attrs
from aiondao.policies import (
    GenerateNewParticipant,
    GenerateNewProposal,
    GenerateNewFunding,
    ActiveProposals,
    ProposalFunding,
    ParticipantVoting,
    ParticipantSellsTokens,
    ParticipantBuysTokens,
    ParticipantExits,
    ParticipantSentiment,
)
from aiondao.utils.networking import bootstrap_network, calc_avg_sentiment
from aiondao.utils._utilz import (
    new_probability_func,
    new_exponential_func,
    new_gamma_func,
    new_random_number_func,
    new_choice_func,
)


def update_collateral_pool(params, step, sL, s, _input):
    commons = s["commons"]
    s["collateral_pool"] = commons._collateral_pool
    return "collateral_pool", commons._collateral_pool


def update_token_supply(params, step, sL, s, _input):
    commons = s["commons"]
    s["token_supply"] = commons._token_supply
    return "token_supply", commons._token_supply


def update_token_price(params, step, sL, s, _input):
    commons = s["commons"]
    s["token_price"] = commons.token_price()
    return "token_price", commons.token_price()


def update_funding_pool(params, step, sL, s, _input):
    commons = s["commons"]
    s["funding_pool"] = commons._funding_pool
    return "funding_pool", commons._funding_pool


def update_avg_sentiment(params, step, sL, s, _input):
    network = s["network"]
    s = calc_avg_sentiment(network)
    return "sentiment", s


def network_deepcopy(params, step, sL, s, _input):
    network = s["network"]
    return "network", copy.deepcopy(network)


def save_policy_output(params, step, sL, s, _input):
    return "policy_output", _input


# This sub-policy block should be run every time the Commons object is updated.
sync_state_variables = {
    "label": "Sync state variables",
    "policies": {},
    "variables": {
        "funding_pool": update_funding_pool,
        "collateral_pool": update_collateral_pool,
        "token_supply": update_token_supply,
        "token_price": update_token_price,
        "sentiment": update_avg_sentiment,
    },
}

network_snapshot = {
    "label": "Get a snapshop of the network",
    "policies": {},
    "variables": {
        "network": network_deepcopy,
    },
}


class CommonsConfig(BaseModel):
    """
    There are just so many options that passing them via kwargs has become
    rather unwieldy. Also it would be useful to have some footnotes next to each
    parameter.

    Inject this paramter into class to control simulation parameters.
    """

    #
    hatch_tribute: float = 0.2
    exit_tribute: float = 0.35
    proposals: int = 2
    hatchers: int = 5
    kappa: int = 2
    vesting_80p_unlocked: int = 120
    days_to_80p_of_max_voting_weight: int = 10
    # maximum fraction of the funding pool that a proposal can ever request
    max_proposal_request: float = 0.2
    timesteps_days: int = 730
    random_seed: Optional[None] = Field(None)
    choice_func: Callable = None

    @root_validator
    def calculate_numerical_functions(cls, values):
        rand_seed = values.get("random_seed", None)
        probability_func = new_probability_func(rand_seed)
        exponential_func = new_exponential_func(rand_seed)
        gamma_func = new_gamma_func(rand_seed)
        random_number_func = new_random_number_func(rand_seed)
        choice_func = new_choice_func(rand_seed)

        values["probability_func"] = probability_func
        values["exponential_func"] = exponential_func
        values["gamma_func"] = gamma_func
        values["random_number_func"] = random_number_func
        values["choice_func"] = choice_func

        vest_unlocked_80 = values.get("vesting_80p_unlocked", 120)

        values["speculation_days"] = int(0.2 * vest_unlocked_80) + int(
            0.6 * vest_unlocked_80 * random_number_func()
        )
        values["multiplier_new_participants"] = 1 + int(9 * random_number_func())

        return values


class CommonsSimulationConfiguration:
    """
    There are just so many options that passing them via kwargs has become
    rather unwieldy. Also it would be useful to have some footnotes next to each
    parameter.
    """

    def __init__(
        self,
        hatchers=5,
        proposals=2,
        hatch_tribute=0.2,
        vesting_80p_unlocked=120,
        exit_tribute=0.35,
        kappa=2,
        days_to_80p_of_max_voting_weight=10,
        max_proposal_request=0.2,
        timesteps_days=730,
        random_seed=None,
    ):
        self.hatchers = hatchers
        self.proposals = proposals
        self.hatch_tribute = hatch_tribute

        # 60 is 2 months until 80% of tokens are unlocked
        self.vesting_80p_unlocked = vesting_80p_unlocked
        self.exit_tribute = exit_tribute
        self.kappa = kappa  # This is the shape of the bonding curve
        self.days_to_80p_of_max_voting_weight = days_to_80p_of_max_voting_weight

        # Proposal may only request up to 20% of the funding pool
        self.max_proposal_request = max_proposal_request

        self.timesteps_days = timesteps_days  # Simulate 2*365=730 days

        self.random_seed = random_seed
        self.probability_func = new_probability_func(random_seed)
        self.exponential_func = new_exponential_func(random_seed)
        self.gamma_func = new_gamma_func(random_seed)
        self.random_number_func = new_random_number_func(random_seed)
        self.choice_func = new_choice_func(random_seed)

        self.speculation_days = int(0.2 * vesting_80p_unlocked) + int(
            0.6 * vesting_80p_unlocked * self.random_number_func()
        )
        self.multiplier_new_participants = 1 + int(9 * self.random_number_func())

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, attrs(self))

    def alpha(self) -> float:
        """
        Converts days_to_80p_of_max_voting_weight to alpha.
        alpha = (1 - 0.8) ^ (1/t)
        """
        return 0.2 ** (1 / self.days_to_80p_of_max_voting_weight)

    def cliff_and_halflife(self) -> Tuple[float, float]:
        """
        This is just a wrapper around hatch.convert_80p_to_cliff_and_halflife().
        It is used here instead for logical consistency (having all derived
        configuration variables come from CommonsSimulationConfiguration), but
        defined in hatch for local consistency (having all hatch-related code in
        one place). Time will tell if this really matters.
        """
        return convert_80p_to_cliff_and_halflife(self.vesting_80p_unlocked)


def bootstrap_simulation(c: CommonsSimulationConfiguration):
    contributions = [c.random_number_func() * 10e5 for i in range(c.hatchers)]
    cliff_days, halflife_days = c.cliff_and_halflife()
    token_batches, initial_token_supply = create_token_batches(
        contributions, 0.1, cliff_days, halflife_days
    )

    commons = Commons(
        sum(contributions),
        initial_token_supply,
        hatch_tribute=c.hatch_tribute,
        exit_tribute=c.exit_tribute,
        kappa=c.kappa,
    )
    network = bootstrap_network(
        token_batches,
        c.proposals,
        commons._funding_pool,
        commons._token_supply,
        c.max_proposal_request,
        c.probability_func,
        c.random_number_func,
        c.gamma_func,
        c.exponential_func,
    )

    initial_conditions = {
        "network": network,
        "commons": commons,
        "funding_pool": commons._funding_pool,
        "collateral_pool": commons._collateral_pool,
        "token_supply": commons._token_supply,
        "token_price": commons.token_price(),
        "policy_output": None,
        "sentiment": 0.75,
    }

    simulation_parameters = {
        "T": range(c.timesteps_days),
        "N": 1,
        "M": {
            # "sentiment_decay": 0.01, #termed mu in the state update function
            # "min_proposal_age_days": 7, # minimum periods passed before a proposal can pass,
            # "sentiment_sensitivity": 0.75,
            # 'min_supp':50, #number of tokens that must be stake for a proposal to be a candidate
            "debug": False,
            "alpha_days_to_80p_of_max_voting_weight": c.alpha(),
            "max_proposal_request": c.max_proposal_request,
            "random_seed": c.random_seed,
            "probability_func": c.probability_func,
            "exponential_func": c.exponential_func,
            "gamma_func": c.gamma_func,
            "random_number_func": c.random_number_func,
            "choice_func": c.choice_func,
            "speculation_days": c.speculation_days,
            "multiplier_new_participants": c.multiplier_new_participants,
        },
    }

    return initial_conditions, simulation_parameters


partial_state_update_blocks = [
    {
        "label": "Generate new participants",
        "policies": {
            "generate_new_participants": GenerateNewParticipant.p_randomly,
        },
        "variables": {
            "network": GenerateNewParticipant.su_add_to_network,
            "commons": GenerateNewParticipant.su_add_investment_to_commons,
        },
    },
    {
        "label": "Update participants' token batch age",
        "policies": {},
        "variables": {
            "network": GenerateNewParticipant.su_update_participants_token_batch_age,
        },
    },
    sync_state_variables,
    {
        "label": "Generate new proposals",
        "policies": {
            "generate_new_proposals": GenerateNewProposal.p_randomly,
        },
        "variables": {
            "network": GenerateNewProposal.su_add_to_network,
        },
    },
    {
        "label": "Generate new funding",
        "policies": {
            "generate_new_funding": GenerateNewFunding.p_exit_tribute_of_average_speculator_position_size,  # TODO: Doesn't interact with the BC
        },
        "variables": {
            "network": GenerateNewFunding.su_add_funding,
        },
    },
    {
        "label": "Update proposals' age and conviction thresholds",
        "policies": {},
        "variables": {
            "network": ProposalFunding.su_update_age_and_conviction_thresholds,
        },
    },
    {
        "label": "Compare proposals' conviction and thresholds",
        "policies": {
            "which_proposals_should_be_funded": ProposalFunding.p_compare_conviction_and_threshold
        },
        "variables": {
            "network": ProposalFunding.su_make_proposal_active,
            "commons": ProposalFunding.su_deduct_funds_from_funding_pool,
            "policy_output": save_policy_output,
        },
    },
    {
        "label": "Update participants' sentiment when a proposal becomes active",
        "policies": {},
        "variables": {
            "network": ParticipantExits.su_update_sentiment_when_proposal_becomes_active,
        },
    },
    sync_state_variables,
    {
        "label": "Proposals become failed or completed",
        "policies": {
            "which_proposals_are_failed_or_completed": ActiveProposals.p_influenced_by_grant_size
        },
        "variables": {
            "network": ActiveProposals.su_set_proposal_status,
            "policy_output": save_policy_output,
        },
    },
    {
        "label": "Update participants' sentiment when a proposal becomes failed or completed",
        "policies": {},
        "variables": {
            "network": ParticipantExits.su_update_sentiment_when_proposal_becomes_failed_or_completed,
        },
    },
    {
        "label": "Participant votes on proposal according to affinity",
        "policies": {
            "participants_stake_tokens_on_proposals": ParticipantVoting.p_participant_votes_on_proposal_according_to_affinity
        },
        "variables": {
            "network": ParticipantVoting.su_update_participants_votes,
        },
    },
    {
        "label": "Calculate proposals' conviction",
        "policies": {},
        "variables": {
            "network": ProposalFunding.su_calculate_conviction,
        },
    },
    {
        "label": "Participant decides to buy tokens",
        "policies": {
            "participants_decide_to_buy_tokens": ParticipantBuysTokens.p_decide_to_buy_tokens_bulk,
        },
        "variables": {
            "commons": ParticipantBuysTokens.su_buy_participants_tokens,
            "network": ParticipantBuysTokens.su_update_participants_tokens,
        },
    },
    sync_state_variables,
    {
        "label": "Participant decides to sell tokens",
        "policies": {
            "participants_decide_to_sell_tokens": ParticipantSellsTokens.p_decide_to_sell_tokens_bulk,
        },
        "variables": {
            "commons": ParticipantSellsTokens.su_burn_participants_tokens,
            "network": ParticipantSellsTokens.su_update_participants_tokens,
        },
    },
    {
        "label": "Participant decides if he wants to exit",
        "policies": {
            "participants_may_exit": ParticipantExits.p_participant_decides_if_he_wants_to_exit,
        },
        "variables": {
            "commons": ParticipantExits.su_burn_exiters_tokens,
            "network": ParticipantExits.su_remove_participants_from_network,
        },
    },
    {
        "label": "Participants' sentiment decays",
        "policies": {},
        "variables": {
            "network": ParticipantSentiment.su_update_sentiment_decay,
        },
    },
    sync_state_variables,
    # network_snapshot,  # Enable it only if running an A/B testing or parameter sweep with a no_deepcopy version of cadCAD
]



from typing import List
from aiondao._imports import *
from aiondao.agents.members import Participant

# isort: on
import random
from aiondao.interfaces.models import IEconomy

# from aiondao.agents.dummy import DummyAgent
from aiondao.agents.commons import (
    CommonAgent,
    create_token_batches,
)

from aiondao.utils.networking import bootstrap_network
from aiondao.libs.states import CommonsConfig, SimState
from aiondao.libs.graph import GraphORM
import networkx as nx


class BootstrapSimulation:
    """Start everything necessary to run the simulation"""

    def get_graph(self, gx: nx.DiGraph):
        return GraphORM(net=gx)

    def bootload(self, confs: CommonsConfig, economy: "IEconomy"):
        state: SimState = self.create_state(confs, economy)
        self.register_participants(state, economy)
        return state

    def register_participants(self, state: SimState, economy: "IEconomy"):
        gx = self.get_graph(state.network)
        participants = gx.parts.all_dict()
        # base_id = random.randint(1200000, 1500000)
        for idx, part in participants.items():
            part.crazy_id = idx + 10001
            part.model = economy
            part.unique_id = idx
            part.logic.reset(state)
            # part.unique_id
            economy.schedule.add(part)
        return economy

    def create_state(self, confs: CommonsConfig, economy: "IEconomy"):
        contributions = [
            confs.random_number_func() * 10e5 for i in range(confs.hatchers)
        ]
        cliff_days, halflife_days = confs.cliff_and_halflife()
        token_batches, initial_token_supply = create_token_batches(
            contributions, 0.1, cliff_days, halflife_days
        )
        anchor_id = random.randint(10000, 12000)
        commons = CommonAgent(
            unique_id=anchor_id,
            model=economy,
            total_hatch_raise=sum(contributions),
            token_supply=initial_token_supply,
            hatch_tribute=confs.hatch_tribute,
            exit_tribute=confs.exit_tribute,
            kappa=confs.kappa,
        )

        # economy.schedule.add(commons)
        network = bootstrap_network(
            token_batches,
            confs.proposals,
            commons._funding_pool,
            commons._token_supply,
            confs.max_proposal_request,
            confs.probability_func,
            confs.random_number_func,
            confs.gamma_func,
            confs.exponential_func,
        )

        return SimState(
            unique_id=anchor_id,
            network=network,
            commons=commons,
            funding_pool=commons._funding_pool,
            collateral_pool=commons._collateral_pool,
            token_supply=commons._token_supply,
            token_price=commons.token_price(),
            policy_output={},
            sentiment=0.75,
            confs=confs,
        )



from aiondao._imports import *


from aiondao.utils.networking import get_participants, get_proposals, ProposalStatus
from aiondao.simulation import CommonsSimulationConfiguration

# from pydantic.dataclasses import dataclass
from auto_all import end_all, start_all

start_all(globals())

# @dataclass
class Metrics(RelaxedModel):
    participants: int
    candidates: int
    funds_candidates: float
    actives: int
    funds_actives: float
    completed: int
    funds_completed: float
    failed: int
    funds_failed: float


# @dataclass(config=)
class CommonsScore(BaseModel, config=autovalid):
    """
    Calculates a final score for a commons simulation run.
    """

    params: CommonsSimulationConfiguration
    df_final: pd.DataFrame
    sigma: float = 130
    sigma_token_price: float = 2
    sigma_funded: float = 5
    penalty: float = -0.2
    final_sentiment_threshold: float = 0.75
    min_sentiment_threshold: float = 0.5
    metrics: Optional[Metrics] = None

    def calc_final_price_ratio(self) -> float:
        """
        Calculates the ratio between final price compared to hatch price
        """
        hatch_price = self.df_final.iloc[0, :]["token_price"]
        final_price = self.df_final.iloc[-1, :]["token_price"]
        final_price_ratio = final_price / (self.sigma_token_price * hatch_price)
        final_result = final_price_ratio
        # Penalty if final price is smaller than 80% of the hatch price
        if final_price < 0.8 * hatch_price:
            final_result += self.penalty
        return min(final_result, 1)

    def calc_avg_price_to_initial_ratio(self) -> float:
        """
        Calculates the average price compared with hatch price
        """
        hatch_price = self.df_final.iloc[0, :]["token_price"]
        avg_price = self.df_final["token_price"].mean()
        avg_price_ratio = (avg_price - hatch_price) / (
            self.sigma_token_price * self.df_final["token_price"].std()
        )
        final_result = avg_price_ratio
        #  Penalty if average price is smaller than hatch price
        if avg_price < hatch_price:
            final_result += self.penalty
        #         return avg_price / hatch_price
        return min(final_result, 1)

    def calc_funded_proposals_ratio(self) -> float:
        """
        Calculates the number of proposals funded compared to
        initial proposals
        """
        init_proposals = self.params.proposals
        #         funded = self.metrics.candidates + self.metrics.actives + self.metrics.completed + self.metrics.failed
        funded = self.metrics.actives + self.metrics.completed + self.metrics.failed
        #         return funded / init_proposals
        funded_proposals_ratio = (funded - init_proposals) / (
            self.sigma_funded * funded
        )
        return min(funded_proposals_ratio, 1)

    def calc_funds_spent_ratio(self) -> float:
        """
        Calculates the total spent by the funding pool compared to
        the amount received in the hatch phase
        """
        hatch_funds = self.df_final.iloc[0, :]["funding_pool"]
        #         total_spent = self.metrics.funds_candidates + self.metrics.funds_actives + self.metrics.funds_completed + self.metrics.funds_failed
        total_spent = (
            self.metrics.funds_actives
            + self.metrics.funds_completed
            + self.metrics.funds_failed
        )
        #         return total_spent / hatch_funds
        funds_spent_ratio = (total_spent - hatch_funds) / total_spent
        return min(funds_spent_ratio, 1)

    def calc_avg_funds_to_initial_ratio(self) -> float:
        """
        Calculates the ration between  average amount in funding pool
        over time compared to the amount received in the hatch phase
        """
        hatch_funds = self.df_final.iloc[0, :]["funding_pool"]
        avg_funds = self.df_final["funding_pool"].mean()
        avg_funds_ratio = avg_funds / hatch_funds
        return min(avg_funds_ratio, 1)

    #         return (avg_funds - hatch_funds) / self.df_final['funding_pool'].std()

    def calc_final_sentiment(self) -> float:
        """
        Obtains the final sentiment at the end of the simulation
        """
        final_sentiment = self.df_final.iloc[-1, :]["sentiment"]
        final_result = final_sentiment
        if final_sentiment < self.final_sentiment_threshold:
            final_result += self.penalty
        return min(final_result, 1)

    def calc_avg_sentiment(self) -> float:
        """
        Obtains the average sentiment
        """
        avg_sentiment = self.df_final["sentiment"].mean()
        return min(avg_sentiment, 1)

    def calc_penalty_min_sentiment(self) -> float:
        """
        If the minimun sentiment smaller than a threshold, add a penalty
        """
        min_sentiment = self.df_final["sentiment"].min()
        if min_sentiment < self.min_sentiment_threshold:
            return self.penalty
        return 0

    def calc_success_to_failed_ratio(self) -> float:
        """
        Calculates the ratio of successful projects to failed ones
        """
        success_to_failed_ratio = self.metrics.completed / self.metrics.failed
        return min(success_to_failed_ratio, 1)

    def calc_participant_to_hatchers_ratio(self) -> float:
        """
        Calculates the ratio between No of final participants and No of hatchers
        """
        hatchers = self.params.hatchers
        participants_ratio = self.metrics.participants / hatchers
        return min(participants_ratio, 1)

    def eval(self) -> float:
        """
        Calculates the final score using all the defined metrics methods in this class
        """
        last_network = self.df_final.iloc[-1, 0]
        p_candidates = get_proposals(last_network, status=ProposalStatus.CANDIDATE)
        candidates = len(p_candidates)
        p_actives = get_proposals(last_network, status=ProposalStatus.ACTIVE)
        actives = len(p_actives)
        p_completed = get_proposals(last_network, status=ProposalStatus.COMPLETED)
        completed = len(p_completed)
        p_failed = get_proposals(last_network, status=ProposalStatus.FAILED)
        failed = len(p_failed)
        participants = len(get_participants(last_network))

        funds_candidates = sum([p.funds_requested for _, p in p_candidates])
        funds_actives = sum([p.funds_requested for _, p in p_actives])
        funds_completed = sum([p.funds_requested for _, p in p_completed])
        funds_failed = sum([p.funds_requested for _, p in p_failed])

        self.metrics = Metrics(
            participants=participants,
            candidates=candidates,
            funds_candidates=funds_candidates,
            actives=actives,
            funds_actives=funds_actives,
            completed=completed,
            funds_completed=funds_completed,
            failed=failed,
            funds_failed=funds_failed,
        )
        methods = [
            attr
            for attr in dir(self)
            if callable(getattr(self, attr)) and attr.startswith("calc_")
        ]
        return min(
            round(sum([getattr(self, method)() for method in methods]) * self.sigma),
            1000,
        )


end_all(globals())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiondao',
 'aiondao.agents',
 'aiondao.interfaces',
 'aiondao.libs',
 'aiondao.policies',
 'aiondao.utils']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'Mesa>=0.9.0,<0.10.0',
 'addict>=2.4.0,<3.0.0',
 'auto-all>=1.4.1,<2.0.0',
 'cadCAD>=0.4.28,<0.5.0',
 'celery[redis]>=5.2.3,<6.0.0',
 'cytoolz>=0.11.2,<0.12.0',
 'decorator>=5.1.1,<6.0.0',
 'devtools>=0.8.0,<0.9.0',
 'fastapi>=0.74.1,<0.75.0',
 'keyring>=23.5.0,<24.0.0',
 'libcst>=0.4.1,<0.5.0',
 'loguru>=0.6.0,<0.7.0',
 'networkx>=2.6.3,<3.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.4.0,<0.5.0',
 'wrapt>=1.13.3,<2.0.0']

entry_points = \
{'console_scripts': ['aiondao = aiondao.cli:app']}

setup_kwargs = {
    'name': 'aiondao',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'kivo360',
    'author_email': 'kivo360@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import typer

from .dev import run_dev

app = typer.Typer()


@app.callback()
def callback():
    """
    Crypto Currency Agent Based Model
    """


@app.command()
def dev():
    """
    Start running the model you want to live.
    """
    typer.secho("Run model", fg="green")
    run_dev()


@app.command()
def start():
    """
    Load the portal gun
    """
    typer.secho("Run model", styles="green")
